from enum import Enum
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy
import torch
from allennlp.common import Registrable
from allennlp.data import Instance
from allennlp.data import Token as AllenNLPToken
from allennlp.data.fields import ListField
from allennlp.data.fields import TextField
from spacy.tokens import Token as SpacyToken

from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.modules.configuration import ComponentConfiguration
from biome.text.modules.heads.task_prediction import Attribution
from biome.text.modules.heads.task_prediction import TaskPrediction
from biome.text.modules.heads.task_prediction import Token

if TYPE_CHECKING:
    from biome.text.configuration import PredictionConfiguration


class TaskName(Enum):
    """The task name enum structure"""

    none = 0
    text_classification = 1
    token_classification = 2
    language_modelling = 3

    def __str__(self) -> str:
        return str(self.name)

    def as_string(self) -> str:
        return str(self)


class TaskHead(torch.nn.Module, Registrable):
    """Base task head class"""

    """The task head name"""
    task_name: TaskName

    def __init__(self, backbone: ModelBackbone):
        super(TaskHead, self).__init__()
        self.backbone = backbone

    def on_vocab_update(self):
        """
        Actions when vocab is updated. Rebuild here modules that initialization depends on some vocab metric

        At this point, the model.vocab is already updated, so it could be used for architecture update
        """
        pass

    @classmethod
    def register(cls, overrides: bool = False, **kwargs):
        """Enables the task head component for pipeline loading"""
        super(TaskHead, TaskHead).register(cls.__name__, exist_ok=overrides)(cls)

    @property
    def labels(self) -> List[str]:
        """The configured vocab labels"""
        return vocabulary.get_labels(self.backbone.vocab)

    @property
    def num_labels(self):
        """The number of vocab labels"""
        return len(self.labels)

    def extend_labels(self, labels: List[str]):
        """Extends the number of labels"""
        vocabulary.extend_labels(self.backbone.vocab, labels)

    def inputs(self) -> Optional[List[str]]:
        """
        The expected inputs names for data featuring. If no defined,
        will be automatically calculated from featurize signature
        """
        return None

    def forward(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """The head's forward pass, it must include the backbone's `forward`.

        When trained, the returned dict has to have a 'loss' key pointing to a
        scalar `torch.Tensor` representing the loss to be optimized.
        When used for inference, it has to include everything to make the TaskOutput -> `self.make_task_output`.
        """
        raise NotImplementedError

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        """Metrics dictionary for training task"""
        raise NotImplementedError

    def featurize(self, *args, **kwargs) -> Instance:
        """Converts incoming data into an Allennlp `Instance`, used for pyTorch tensors generation

        Returns
        -------
        instance

        Raises
        ------
        FeaturizeError
        """
        raise NotImplementedError

    def make_task_prediction(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
        prediction_config: "PredictionConfiguration",
    ) -> TaskPrediction:
        """Transforms the forward output to a task output, only used for predictions.

        Parameters
        ----------
        single_forward_output
            A single (not batched) output from the head's forward method
        instance
            The instance underlying the prediction
        prediction_config
            Configurations for the prediction

        Returns
        -------
        task_prediction
            A task specific output for the prediction
        """
        prediction = self._make_task_prediction(single_forward_output, instance)

        if prediction_config.add_tokens:
            prediction.tokens = self._extract_tokens(instance)

        if prediction_config.add_attributions:
            prediction.attributions = self._compute_attributions(
                single_forward_output, instance, **prediction_config.attributions_kwargs
            )

        return prediction

    def _make_task_prediction(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
    ) -> TaskPrediction:
        """Makes a basic task prediction.

        Must be implemented by the child class.

        Parameters
        ----------
        single_forward_output
            A single (not batched) output from the head's forward method
        instance
            The instance underlying the prediction

        Returns
        -------
        task_prediction
            A task specific output for the prediction
        """
        # One could implement a generic solution to just pass on the forward_output, but it would be slow and i
        # recommend thinking about what a prediction should return, it is very likely not the same as for the forward.
        # Possible solution:
        # Dynamically create a dataclass with necessary fields: C = dataclasses.make_dataclass(...)
        # Inherit from TaskPrediction: return type("...", (C, TaskPrediction, ), {})(**forward_output)
        raise NotImplementedError("Predictions are not implemented in this head")

    def _extract_tokens(self, instance: Instance) -> List[Union[Token, List[Token]]]:
        """Extracts the tokens from all TextFields in an instance.

        This is a generic implementation and you might have to overwrite it for your specific head.

        Parameters
        ----------
        instance
            The instance underlying the prediction

        Returns
        -------
        tokens
        """
        tokens: List[Union[Token, List[Token]]] = []

        for field_name, field in instance.items():
            if isinstance(field, TextField):
                tokens += self._extract_tokens_from_text_field(field, field_name)
            elif isinstance(field, ListField):
                for single_field in field:
                    if isinstance(single_field, TextField):
                        tokens.append(
                            self._extract_tokens_from_text_field(
                                single_field, field_name
                            )
                        )

        return tokens

    def _extract_tokens_from_text_field(
        self, field: TextField, name: str
    ) -> List[Token]:
        """Helper function for `self._extract_tokens`"""
        return [
            Token(
                text=token.text,
                start=token.idx,
                end=self._get_token_end(token),
                field=name,
            )
            for token in field
        ]

    @staticmethod
    def _get_token_end(token: Union[AllenNLPToken, SpacyToken]) -> Optional[int]:
        """While AllenNLP tokens have an idx_end, spacy Tokens do not"""
        try:
            return token.idx_end
        except AttributeError:
            return token.idx + len(token.text) if isinstance(token.idx, int) else None

    def _compute_attributions(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
        **kwargs
    ) -> List[Union[Attribution, List[Attribution]]]:
        """Tries to attribute the prediction to input features.

        Must be implemented by the child class.

        Parameters
        ----------
        single_forward_output
            A single (not batched) output from the head's forward method
        instance
            The instance underlying the prediction

        Returns
        -------
        attributions
        """
        raise NotImplementedError(
            "Attributing the prediction to the input is not implemented in this head"
        )


class TaskHeadConfiguration(ComponentConfiguration[TaskHead]):
    """Layer spec for TaskHead components"""

    pass



import pytest

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import VocabularyConfiguration
from biome.text.errors import EmptyVocabError
from biome.text.features import CharFeatures
from biome.text.features import TransformersFeatures
from biome.text.features import WordFeatures


@pytest.fixture
def pipeline():
    config = {
        "name": "vocab_test",
        "features": {
            "transformers": {"model_name": "sshleifer/tiny-distilbert-base-cased"},
            "word": {"embedding_dim": 2},
            "char": {
                "embedding_dim": 2,
                "dropout": 0.1,
                "encoder": {
                    "type": "gru",
                    "hidden_size": 2,
                    "num_layers": 1,
                    "bidirectional": False,
                },
            },
        },
        "head": {
            "type": "TextClassification",
            "labels": ["good", "bad"],
        },
    }

    return Pipeline.from_config(config)


@pytest.fixture
def train_dataset():
    data = {"text": ["this is a test", "and another one"], "label": ["good", "bad"]}
    return Dataset.from_dict(data)


@pytest.fixture
def valid_dataset():
    data = {
        "text": ["and what about the validation", "do not forget this one"],
        "label": ["bad", "good"],
    }
    return Dataset.from_dict(data)


def test_default_vocab(pipeline, train_dataset, valid_dataset):
    # Transformer vocab is added on pipeline creation
    assert pipeline.vocab.get_vocab_size(TransformersFeatures.namespace) == 28996
    # While word and char vocab should be empty (except for the oov and padding token)
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 2
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 2

    # Training should build a default vocab with only the training dataset
    Trainer(pipeline, train_dataset=train_dataset)
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 9
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 12
    assert pipeline.vocab.get_vocab_size(TransformersFeatures.namespace) == 28996

    # Pretrained pipelines should extend the vocab by default
    Trainer(pipeline, train_dataset=valid_dataset)
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 16
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 19
    assert pipeline.vocab.get_vocab_size(TransformersFeatures.namespace) == 28996


def test_specific_vocab_config(pipeline, train_dataset, valid_dataset):
    vocab_config = VocabularyConfiguration(include_valid_data=True)

    Trainer(
        pipeline,
        train_dataset=train_dataset,
        valid_dataset=valid_dataset,
        vocab_config=vocab_config,
    )
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 16
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 19
    assert pipeline.vocab.get_vocab_size(TransformersFeatures.namespace) == 28996


def test_not_touching_vocab(pipeline, train_dataset, valid_dataset):
    # vocab_config=None leaves the pipeline's vocab empty from an unpretrained pipeline
    with pytest.raises(EmptyVocabError):
        Trainer(pipeline, train_dataset=train_dataset, vocab_config=None)

    # vocab_config=None should not extend the vocab for a pretrained pipeline
    Trainer(pipeline, train_dataset=train_dataset)
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 9
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 12
    Trainer(pipeline, train_dataset=valid_dataset, vocab_config=None)
    assert pipeline.vocab.get_vocab_size(WordFeatures.namespace) == 9
    assert pipeline.vocab.get_vocab_size(CharFeatures.namespace) == 12



from typing import Dict

import pytest

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import TrainerConfiguration


@pytest.fixture
def training_dataset() -> Dataset:
    """Creating the dataframe."""
    data = {
        "text": [
            "The most common audits were about waste and recycling.",
            "The company fabricates plastic chairs.",
        ],
        "entities": [
            [
                {"start": 34, "end": 39, "label": "PN", "text": "waste"},
                {"start": 16, "end": 22, "label": "QTY", "text": "audits"},
            ],
            [
                {"start": 4, "end": 11, "label": "OBJECT", "text": "company"},
                {"start": 31, "end": 37, "label": "SUBJECT", "text": "chairs"},
            ],
        ],
        "label": ["Message-Topic(e1,e2)", "Product-Producer(e2,e1)"],
    }

    return Dataset.from_dict(data)


@pytest.fixture
def pipeline_dict() -> Dict:
    """Creating the pipeline dictionary"""

    pipeline_dict = {
        "name": "biome-rele",
        "features": {
            "word": {"embedding_dim": 2},
            "char": {
                "embedding_dim": 2,
                "dropout": 0.1,
                "encoder": {
                    "type": "gru",
                    "hidden_size": 2,
                },
            },
        },
        "head": {
            "type": "RelationClassification",
            "labels": ["Message-Topic(e1,e2)", "Product-Producer(e2,e1)"],
            "entities_embedder": {"num_embeddings": 12, "embedding_dim": 50},
            "feedforward": {
                "num_layers": 1,
                "hidden_dims": [4],
                "activations": ["relu"],
                "dropout": [0.1],
            },
        },
    }

    return pipeline_dict


@pytest.fixture
def trainer_config() -> TrainerConfiguration:
    return TrainerConfiguration(
        max_epochs=1,
        optimizer={"type": "adamw", "lr": 0.002},
        gpus=0,
    )


def test_train(pipeline_dict, training_dataset, trainer_config, tmp_path):
    """Testing a classifier made from scratch"""

    pipeline = Pipeline.from_config(pipeline_dict)
    pipeline.predict(
        text="The most common audits were about waste and recycling",
        entities=[
            {"start": 34, "end": 39, "label": "OBJECT", "text": "waste"},
            {"start": 16, "end": 22, "label": "SUBJECT", "text": "audits"},
        ],
    )

    trainer = Trainer(
        pipeline=pipeline,
        train_dataset=training_dataset,
        valid_dataset=training_dataset,
        trainer_config=trainer_config,
    )
    trainer.fit(tmp_path / "relation_classifier")

    # test loading
    Pipeline.from_pretrained(tmp_path / "relation_classifier" / "model.tar.gz")



import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy
import torch
from allennlp.data import Instance
from allennlp.data.fields import LabelField
from allennlp.data.fields import MultiLabelField

from biome.text import helpers
from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.featurizer import FeaturizeError
from biome.text.metrics import Metrics
from biome.text.modules.heads.task_head import TaskHead
from biome.text.modules.heads.task_head import TaskName
from biome.text.modules.heads.task_prediction import Attribution
from biome.text.modules.heads.task_prediction import TaskPrediction


class ClassificationHead(TaskHead):
    """Base abstract class for classification problems

    Parameters
    ----------
    labels
        A list of labels for your classification task
    multilabel
        Is this a multi label classification task? Default: False
    label_weights
        A list of weights for each label. The weights must be in the same order as the `labels`.
        You can also provide a dictionary that maps the label to its weight. Default: None.
    """

    task_name = TaskName.text_classification
    _LOGGER = logging.getLogger(__name__)

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        multilabel: bool = False,
        label_weights: Optional[Union[List[float], Dict[str, float]]] = None,
    ):
        super().__init__(backbone)
        vocabulary.set_labels(self.backbone.vocab, labels)

        # label related configurations
        self._multilabel = multilabel

        # metrics and loss
        if isinstance(label_weights, list):
            label_weights = torch.tensor(label_weights, dtype=torch.float32)
        elif isinstance(label_weights, dict):
            label_weights = torch.tensor(
                [label_weights[label] for label in labels], dtype=torch.float32
            )
        if self._multilabel:
            self._loss = torch.nn.BCEWithLogitsLoss(weight=label_weights)
            self._metrics = Metrics(
                micro={"type": "fbeta_multi_label", "average": "micro"},
                macro={"type": "fbeta_multi_label", "average": "macro"},
                per_label={
                    "type": "fbeta_multi_label",
                    "labels": [i for i in range(len(labels))],
                },
            )
        else:
            self._loss = torch.nn.CrossEntropyLoss(weight=label_weights)
            self._metrics = Metrics(
                accuracy={"type": "categorical_accuracy"},
                micro={"type": "fbeta", "average": "micro"},
                macro={"type": "fbeta", "average": "macro"},
                per_label={"type": "fbeta", "labels": [i for i in range(len(labels))]},
            )

    def _add_label(
        self,
        instance: Instance,
        label: Union[List[str], List[int], str, int],
        to_field: str = "label",
    ) -> Instance:
        """Adds the label field for classification into the instance data

        Helper function for the child's `self.featurize` method.

        Parameters
        ----------
        instance
            Add a label field to this instance
        label
            The label data
        to_field
            Name space of the field

        Returns
        -------
        instance
            If `label` is not None, return `instance` with the a label field added.
            Otherwise return just the given `instance`.

        Raises
        ------
        FeaturizeError
            If the label is an empty string or does not match the type:
            - (str, int) for single label
            - (list, np.array) for multi label
        """
        # "if not label:" fails for ndarrays, this is why we explicitly check for None
        if label is None:
            return instance

        field = None
        # check if multilabel and if adequate type
        if self._multilabel and isinstance(label, (list, numpy.ndarray)):
            label = label.tolist() if isinstance(label, numpy.ndarray) else label
            field = MultiLabelField(label, label_namespace=vocabulary.LABELS_NAMESPACE)
        # check if not multilabel and adequate type + check for empty strings
        if not self._multilabel and isinstance(label, (str, int)) and label:
            field = LabelField(label, label_namespace=vocabulary.LABELS_NAMESPACE)
        if not field:
            # We have label info but we cannot build the label field --> discard the instance
            raise FeaturizeError(f"Cannot create label field for `label={label}`!")

        instance.add_field(to_field, field)

        return instance

    def _make_forward_output(
        self, logits: torch.Tensor, label: Optional[torch.IntTensor]
    ) -> Dict[str, Any]:
        """Returns a dict with the logits and optionally the loss

        Helper function for the child's `self.forward` method.
        """
        if label is not None:
            return {
                "loss": self._compute_metrics_and_return_loss(logits, label),
                "logits": logits,
            }

        return {"logits": logits}

    def _compute_metrics_and_return_loss(
        self, logits: torch.Tensor, label: torch.IntTensor
    ) -> float:
        """Helper function for the `self._make_forward_output` method."""
        for metric in self._metrics.get_dict(is_train=self.training).values():
            metric(logits, label)

        if self._multilabel:
            # casting long to float for BCELoss
            # see https://discuss.pytorch.org/t/nn-bcewithlogitsloss-cant-accept-one-hot-target/59980
            return self._loss(
                logits.view(-1, self.num_labels),
                label.view(-1, self.num_labels).type_as(logits),
            )

        return self._loss(logits, label.long())

    def _compute_labels_and_probabilities(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
    ) -> Tuple[List[str], List[float]]:
        """Computes the probabilities based on the logits and looks up the labels

        This is a helper function for the `self._make_task_prediction` of the children.

        Parameters
        ----------
        single_forward_output
            A single (not batched) output from the head's forward method

        Returns
        -------
        (labels, probabilities)
        """
        logits = torch.from_numpy(single_forward_output["logits"])

        if self._multilabel:
            probabilities = logits.sigmoid()
        else:
            probabilities = torch.nn.functional.softmax(logits, dim=0)

        labels, all_probabilities = (
            self._add_and_sort_labels_and_probabilities(probabilities)
            if self.num_labels > 0
            else ([], [])
        )

        return labels, all_probabilities

    def _add_and_sort_labels_and_probabilities(
        self, probabilities: torch.Tensor
    ) -> Tuple[List[str], List[float]]:
        """Returns the labels and probabilities sorted by the probability (descending)

        Helper function for the `self._compute_labels_and_probabilities` method. The list of the returned
        probabilities can be larger than the input probabilities, since we add all defined labels in the head.

        Parameters
        ----------
        probabilities
            Probabilities of the model's prediction for one instance

        Returns
        -------
        labels, probabilities
        """
        all_classes_probs = torch.zeros(
            self.num_labels,  # this can be >= probabilities.size()[0]
            device=probabilities.get_device()
            if probabilities.get_device() > -1
            else None,
        )
        all_classes_probs[: probabilities.size()[0]] = probabilities
        sorted_indexes_by_prob = torch.argsort(
            all_classes_probs, descending=True
        ).tolist()

        labels = [
            vocabulary.label_for_index(self.backbone.vocab, idx)
            for idx in sorted_indexes_by_prob
        ]
        probabilities = [
            float(all_classes_probs[idx]) for idx in sorted_indexes_by_prob
        ]

        return labels, probabilities

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        """Get the metrics of our classifier, see :func:`~allennlp_2.models.Model.get_metrics`.

        Parameters
        ----------
        reset
            Reset the metrics after obtaining them?

        Returns
        -------
        A dictionary with all metric names and values.
        """
        metrics, final_metrics = self._metrics.get_dict(is_train=self.training), {}
        for name, metric in metrics.items():
            if name == "accuracy":
                final_metrics.update({"accuracy": metric.get_metric(reset)})
            elif name in ["macro", "micro"]:
                final_metrics.update(
                    {
                        f"{name}/{key}": value
                        for key, value in metric.get_metric(reset).items()
                    }
                )
            elif name == "per_label":
                for key, values in metric.get_metric(reset).items():
                    for i, value in enumerate(values):
                        label = vocabulary.label_for_index(self.backbone.vocab, i)
                        # sanitize label using same patterns as tensorboardX to avoid summary writer warnings
                        label = helpers.sanitize_metric_name(label)
                        final_metrics.update({f"_{key}/{label}": value})

        return final_metrics

    def forward(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        raise NotImplementedError

    def featurize(self, *args, **kwargs) -> Instance:
        raise NotImplementedError

    def _make_task_prediction(
        self, single_forward_output: Dict[str, numpy.ndarray], instance: Instance
    ) -> TaskPrediction:
        raise NotImplementedError

    def _compute_attributions(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
        **kwargs,
    ) -> List[Union[Attribution, List[Attribution]]]:
        raise NotImplementedError



import pytest
from numpy.testing import assert_allclose

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import TrainerConfiguration


@pytest.fixture
def pipeline():
    return Pipeline.from_config(
        {
            "name": "test_pipeline_copy",
            "head": {
                "type": "TextClassification",
                "labels": ["a", "b"],
            },
        }
    )


@pytest.fixture
def dataset():
    return Dataset.from_dict(
        {
            "text": ["this is", "a test"],
            "label": ["a", "b"],
        }
    )


def test_copy(pipeline):
    prediction = pipeline.predict("check this")
    pipeline_copy = pipeline.copy()
    prediction_copy = pipeline_copy.predict("check this")

    assert_allclose(prediction["probabilities"], prediction_copy["probabilities"])


def test_train_from_pretrained(pipeline, dataset, tmp_path):
    output_path = tmp_path / "test_train_from_pretrained_output"
    trainer_config = TrainerConfiguration(max_epochs=1, batch_size=2, gpus=0)
    trainer = Trainer(
        pipeline=pipeline, train_dataset=dataset, trainer_config=trainer_config
    )
    trainer.fit(output_path)

    prediction = pipeline.predict("a test")
    pipeline_loaded = Pipeline.from_pretrained(output_path / "model.tar.gz")
    prediction_loaded = pipeline_loaded.predict("a test")

    assert_allclose(prediction["probabilities"], prediction_loaded["probabilities"])



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from typing import cast

import numpy
import torch
from allennlp.data import Instance
from allennlp.data import TextFieldTensors
from allennlp.data.fields import ListField
from allennlp.data.fields import TextField
from allennlp.modules.seq2seq_encoders import PassThroughEncoder
from allennlp.modules.seq2vec_encoders import BagOfEmbeddingsEncoder
from allennlp.nn.util import get_text_field_mask
from captum.attr import IntegratedGradients

from biome.text.backbone import ModelBackbone
from biome.text.modules.configuration import ComponentConfiguration
from biome.text.modules.configuration import FeedForwardConfiguration
from biome.text.modules.configuration import Seq2SeqEncoderConfiguration
from biome.text.modules.configuration import Seq2VecEncoderConfiguration
from biome.text.modules.encoders import TimeDistributedEncoder
from biome.text.modules.heads.classification.classification import ClassificationHead
from biome.text.modules.heads.task_prediction import Attribution
from biome.text.modules.heads.task_prediction import DocumentClassificationPrediction


class DocumentClassification(ClassificationHead):
    """
    Task head for document text classification. It's quite similar to text
    classification but including the doc2vec transformation layers

    Parameters
    ----------
    backbone
        The backbone of your model. Must not be provided when initiating with `Pipeline.from_config`.
    labels
        A list of labels for your classification task.
    token_pooler
        The pooler at token level to provide one vector per sentence. Default: `BagOfEmbeddingsEncoder`.
    sentence_encoder
        An optional sequence to sequence encoder that contextualizes the sentence representations. Default: None.
    sentence_pooler
        The pooler at sentence level to provide a vector for the document. Default: `BagOfEmbeddingsEncoder`.
    feedforward
        An optional feedforward layer applied to the output of the sentence pooler. Default: None.
    multilabel
        Is this a multi label classification task? Default: False
    label_weights
        A list of weights for each label. The weights must be in the same order as the `labels`.
        You can also provide a dictionary that maps the label to its weight. Default: None.
    """

    forward_arg_name = "text"
    label_name = "label"

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        token_pooler: Optional[Seq2VecEncoderConfiguration] = None,
        sentence_encoder: Optional[Seq2SeqEncoderConfiguration] = None,
        sentence_pooler: Seq2VecEncoderConfiguration = None,
        feedforward: Optional[FeedForwardConfiguration] = None,
        multilabel: bool = False,
        label_weights: Optional[Union[List[float], Dict[str, float]]] = None,
    ) -> None:

        super().__init__(
            backbone,
            labels=labels,
            multilabel=multilabel,
            label_weights=label_weights,
        )

        self.backbone.encoder = TimeDistributedEncoder(backbone.encoder)

        # layers
        self.token_pooler = TimeDistributedEncoder(
            BagOfEmbeddingsEncoder(embedding_dim=self.backbone.encoder.get_output_dim())
            if not token_pooler
            else token_pooler.input_dim(
                self.backbone.encoder.get_output_dim()
            ).compile()
        )
        self.sentence_encoder = (
            PassThroughEncoder(self.token_pooler.get_output_dim())
            if not sentence_encoder
            else sentence_encoder.input_dim(
                self.token_pooler.get_output_dim()
            ).compile()
        )
        self.sentence_pooler = (
            BagOfEmbeddingsEncoder(self.sentence_encoder.get_output_dim())
            if not sentence_pooler
            else sentence_pooler.input_dim(
                self.sentence_encoder.get_output_dim()
            ).compile()
        )
        self.feedforward = (
            None
            if not feedforward
            else feedforward.input_dim(self.sentence_pooler.get_output_dim()).compile()
        )

        self._classification_layer = torch.nn.Linear(
            (self.feedforward or self.sentence_pooler).get_output_dim(),
            self.num_labels,
        )

    def featurize(
        self,
        text: Union[List[str], Dict[str, str]],
        label: Optional[Union[str, List[str]]] = None,
    ) -> Instance:
        instance = self.backbone.featurizer(
            text, to_field=self.forward_arg_name, exclude_record_keys=True
        )

        return self._add_label(instance, label, to_field=self.label_name)

    def forward(
        self, text: TextFieldTensors, label: torch.IntTensor = None
    ) -> Dict[str, Any]:
        mask = get_text_field_mask(text, num_wrapping_dims=1)
        embeddings = self.backbone.embedder(text, num_wrapping_dims=1)
        logits = self._encoder_and_head_forward(embeddings, mask)

        output = self._make_forward_output(logits, label)

        # For computing the attributions
        # TODO: An optimized implementation would be to calculate the attributions directly in the forward method
        #  and provide a practical switch, maybe: `with head.turn_attributions_on(): self.forward_on_instances()`
        #  In this way we would calculate the attributions batch wise and on on GPU if available.
        if not self.training:
            output["embeddings"], output["mask"] = embeddings, mask

        return output

    def _encoder_and_head_forward(
        self, embeddings: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """We reuse this method for the computation of the attributions"""
        encoded_text = self.token_pooler(
            self.backbone.encoder(embeddings, mask=mask), mask=mask
        )

        # Here we need to mask the TextFields that only contain the padding token -> last dimension only contains False
        # Those fields were added to possibly equalize the batch.
        mask = torch.sum(mask, -1) > 0
        encoded_text = self.sentence_pooler(
            self.sentence_encoder(encoded_text, mask=mask), mask=mask
        )

        if self.feedforward is not None:
            encoded_text = self.feedforward(encoded_text)

        return self._classification_layer(encoded_text)

    def _compute_attributions(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
        n_steps: int = 50,
    ) -> List[List[Attribution]]:
        """Attributes the prediction to the input.

        The attributions are calculated by means of the [Integrated Gradients](https://arxiv.org/abs/1703.01365) method.

        Parameters
        ----------
        single_forward_output
            Non-batched forward output containing numpy arrays
        instance
            The instance containing the input data
        n_steps
            The number of steps used when calculating the attribution of each token.

        Returns
        -------
        attributions
            A list of list of attributions due to the the ListField level
        """
        # captum needs `torch.Tensor`s and we need a batch dimension (-> unsqueeze)
        embeddings = torch.from_numpy(single_forward_output["embeddings"]).unsqueeze(0)
        mask = torch.from_numpy(single_forward_output["mask"]).unsqueeze(0)
        logits = torch.from_numpy(single_forward_output["logits"]).unsqueeze(0)

        ig = IntegratedGradients(self._encoder_and_head_forward)
        attributions, delta = ig.attribute(
            embeddings,
            n_steps=n_steps,
            target=torch.argmax(logits),
            additional_forward_args=mask,
            return_convergence_delta=True,
        )
        attributions = attributions.sum(dim=3).squeeze(0)
        attributions = attributions / torch.norm(attributions)
        attributions = attributions.detach().numpy()

        document_tokens = [
            cast(TextField, text_field).tokens
            for text_field in cast(ListField, instance.get(self.forward_arg_name))
        ]

        return [
            [
                Attribution(
                    text=token.text,
                    start=token.idx,
                    end=self._get_token_end(token),
                    field=self.forward_arg_name,
                    attribution=attribution,
                )
                for token, attribution in zip(sentence_tokens, sentence_attributions)
            ]
            for sentence_tokens, sentence_attributions in zip(
                document_tokens, attributions
            )
        ]

    def _make_task_prediction(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
    ) -> DocumentClassificationPrediction:
        labels, probabilities = self._compute_labels_and_probabilities(
            single_forward_output
        )

        return DocumentClassificationPrediction(
            labels=labels, probabilities=probabilities
        )


class DocumentClassificationConfiguration(
    ComponentConfiguration[DocumentClassification]
):
    """Lazy initialization for document classification head components"""

    pass



import copy
import logging
import os
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import yaml
from allennlp.common import FromParams
from allennlp.common import Params
from allennlp.common.checks import ConfigurationError
from allennlp.data import TokenIndexer
from allennlp.data import Vocabulary
from allennlp.modules import TextFieldEmbedder
from pytorch_lightning import Callback
from pytorch_lightning.loggers import LightningLoggerBase

from biome.text.features import CharFeatures
from biome.text.features import TransformersFeatures
from biome.text.features import WordFeatures
from biome.text.featurizer import InputFeaturizer
from biome.text.helpers import sanitize_for_params
from biome.text.helpers import sanitize_for_yaml
from biome.text.helpers import save_dict_as_yaml
from biome.text.modules.encoders import Encoder
from biome.text.modules.heads.classification.relation_classification import (
    RelationClassification,
)
from biome.text.modules.heads.task_head import TaskHeadConfiguration
from biome.text.modules.heads.token_classification import TokenClassification
from biome.text.tokenizer import Tokenizer
from biome.text.tokenizer import TransformersTokenizer

_LOGGER = logging.getLogger(__name__)


class FeaturesConfiguration(FromParams):
    """Configures the input features of the `Pipeline`

    Use this for defining the features to be used by the model, namely word and character embeddings.

    :::tip
    If you do not pass in either of the parameters (`word` or `char`),
    your pipeline will be setup with a default word feature (embedding_dim=50).
    :::

    Example:

    ```python
    word = WordFeatures(embedding_dim=100)
    char = CharFeatures(embedding_dim=16, encoder={'type': 'gru'})
    config = FeaturesConfiguration(word, char)
    ```

    Parameters
    ----------
    word
        The word feature configurations, see `biome.text.features.WordFeatures`
    char
        The character feature configurations, see `biome.text.features.CharFeatures`
    transformers
        The transformers feature configuration, see `biome.text.features.TransformersFeatures`
        A word-level representation of the [transformer](https://huggingface.co/models) models using AllenNLP's

    """

    __DEFAULT_CONFIG = WordFeatures(embedding_dim=50)

    def __init__(
        self,
        word: Optional[WordFeatures] = None,
        char: Optional[CharFeatures] = None,
        transformers: Optional[TransformersFeatures] = None,
    ):
        self.word = word
        self.char = char
        self.transformers = transformers

        if not (word or char or transformers):
            self.word = self.__DEFAULT_CONFIG

    @classmethod
    def from_params(
        cls: Type["FeaturesConfiguration"], params: Params, **extras
    ) -> "FeaturesConfiguration":

        word = params.pop("word", None)
        word = WordFeatures(**word.as_dict(quiet=True)) if word else None

        char = params.pop("char", None)
        char = CharFeatures(**char.as_dict(quiet=True)) if char else None

        transformers = params.pop("transformers", None)
        transformers = (
            TransformersFeatures(**transformers.as_dict(quiet=True))
            if transformers
            else None
        )

        params.assert_empty("FeaturesConfiguration")
        return cls(word=word, char=char, transformers=transformers)

    @property
    def configured_namespaces(self) -> List[str]:
        """Return the namespaces of the features that are configured"""
        return [feature.namespace for feature in self]

    def compile_embedder(self, vocab: Vocabulary) -> TextFieldEmbedder:
        """Creates the embedder based on the configured input features

        Parameters
        ----------
        vocab
            The vocabulary for which to create the embedder

        Returns
        -------
        embedder
        """
        configuration = self._make_allennlp_config()

        # We have to set the weights_file / pretrained_file (pretrained word vectors) to None for two reasons:
        # - compiling the embedder with a weights file and an empty vocab will fail,
        #   this is the case for untrained pipelines
        # - compiling the embedder with a non existent weights file will fail,
        #   but for pretrained pipelines it should be optional
        # We will "reactivate" it after the embedder is initialized for extending the vocab with it.
        try:
            configuration["word"]["embedder"]["pretrained_file"] = None
        except KeyError:
            pass

        text_field_embedder = TextFieldEmbedder.from_params(
            Params(
                {
                    "token_embedders": {
                        feature_namespace: config["embedder"]
                        for feature_namespace, config in configuration.items()
                    }
                }
            ),
            vocab=vocab,
        )

        # It is save to reactivate it, since extending the vocab will only throw a warning if the file does not exist.
        if self.word is not None:
            setattr(
                getattr(text_field_embedder, f"token_embedder_{self.word.namespace}"),
                "_pretrained_file",
                self.word.weights_file,
            )

        return text_field_embedder

    def compile_featurizer(self, tokenizer: Tokenizer) -> InputFeaturizer:
        """Creates the featurizer based on the configured input features

        :::tip
        If you are creating configurations programmatically
        use this method to check that you provided a valid configuration.
        :::

        Parameters
        ----------
        tokenizer
            Tokenizer used for this featurizer

        Returns
        -------
        featurizer
            The configured `InputFeaturizer`
        """
        configuration = self._make_allennlp_config()

        indexer = {
            feature_namespace: TokenIndexer.from_params(Params(config["indexer"]))
            for feature_namespace, config in configuration.items()
        }

        return InputFeaturizer(tokenizer, indexer=indexer)

    def _make_allennlp_config(self) -> Dict[str, Any]:
        """Returns a configuration dict compatible with allennlp

        Returns
        -------
        config_dict
        """
        configuration = {feature.namespace: feature.config for feature in self}

        return copy.deepcopy(configuration)

    def __iter__(self):
        for feature in [self.word, self.char, self.transformers]:
            if feature is not None:
                yield feature


class TokenizerConfiguration(FromParams):
    """Configures the `Tokenizer`

    Parameters
    ----------
    spacy_model
        The [spaCy model](https://spacy.io/models) used for the tokenization. Default: "en_core_web_sm".
    max_sequence_length
        Maximum length in characters for input texts truncated with `[:max_sequence_length]` after `TextCleaning`.
    max_nr_of_sentences
        Maximum number of sentences to keep when using `segment_sentences` truncated with `[:max_sequence_length]`.
    text_cleaning
        A `TextCleaning` configuration with pre-processing rules for cleaning up and transforming raw input text.
    segment_sentences
        Whether to segment input texts into sentences.
    use_spacy_tokens
        If True, the tokenized token list contains spacy tokens instead of allennlp tokens
    remove_space_tokens
        If True, all found space tokens will be removed from the final token list.
    start_tokens
        A list of token strings to the sequence before tokenized input text.
    end_tokens
        A list of token strings to the sequence after tokenized input text.
    use_transformers
        If true, we will use a transformers tokenizer from HuggingFace and disregard all other parameters above
        (except `max_sequence_length`!).
        If you specify any of the above parameters you want to set this to false.
        If None, we automatically choose the right value based on your feature and head configuration.
    transformers_kwargs
        This dict is passed on to AllenNLP's `PretrainedTransformerTokenizer`.
        If no `model_name` key is provided, we will infer one from the features configuration.
    """

    # note: It's important that it inherits from FromParas so that `Pipeline.from_pretrained()` works!
    def __init__(
        self,
        spacy_model: str = "en_core_web_sm",
        max_sequence_length: int = None,
        max_nr_of_sentences: int = None,
        text_cleaning: Optional[Dict[str, Any]] = None,
        segment_sentences: bool = False,
        use_spacy_tokens: bool = False,
        remove_space_tokens: bool = True,
        start_tokens: Optional[List[str]] = None,
        end_tokens: Optional[List[str]] = None,
        use_transformers: Optional[bool] = None,
        transformers_kwargs: Optional[Dict] = None,
    ):
        self.spacy_model = spacy_model
        self.max_sequence_length = max_sequence_length
        self.max_nr_of_sentences = max_nr_of_sentences
        self.segment_sentences = segment_sentences
        self.text_cleaning = text_cleaning
        self.start_tokens = start_tokens
        self.end_tokens = end_tokens
        self.use_spacy_tokens = use_spacy_tokens
        self.remove_space_tokens = remove_space_tokens
        self.use_transformers = use_transformers
        self.transformers_kwargs = transformers_kwargs or {}

    def __eq__(self, other):
        return all(a == b for a, b in zip(vars(self), vars(other)))


class PipelineConfiguration(FromParams):
    """Creates a `Pipeline` configuration

    Parameters
    ----------
    name
        The `name` for our pipeline
    features
        The input `features` to be used by the model pipeline. We define this using a `FeaturesConfiguration` object.
    head
        The `head` for the task, e.g., a LanguageModelling task, using a `TaskHeadConfiguration` object.
    tokenizer
        The `tokenizer` defined with a `TokenizerConfiguration` object.
    encoder
        The core text seq2seq `encoder` of our model using a `Seq2SeqEncoderConfiguration`
    """

    # To be able to skip the configuration checks when testing
    _SKIP_CHECKS = False

    def __init__(
        self,
        name: str,
        head: TaskHeadConfiguration,
        features: Optional[FeaturesConfiguration] = None,
        tokenizer: Optional[TokenizerConfiguration] = None,
        encoder: Optional[Encoder] = None,
    ):
        super(PipelineConfiguration, self).__init__()

        self.name = name
        self.head = head
        self.features = features or FeaturesConfiguration()
        self.tokenizer_config = tokenizer or TokenizerConfiguration()

        # figure out if we need a transformers tokenizer
        if self.tokenizer_config.use_transformers is None:
            self._use_transformers_tokenizer_if_sensible()

        # make sure we use the right indexer/embedder for the transformers feature
        if self.tokenizer_config.use_transformers:
            self.features.transformers.mismatched = False
            if self.tokenizer_config.transformers_kwargs.get("model_name") is None:
                self.tokenizer_config.transformers_kwargs[
                    "model_name"
                ] = self.features.transformers.model_name

        if not self._SKIP_CHECKS:
            self._check_for_incompatible_configurations()

        self.encoder = encoder

    def _use_transformers_tokenizer_if_sensible(self):
        if (
            # Only use word pieces if no word-based feature was chosen
            self.features.transformers is not None
            and self.features.word is None
            and self.features.char is None
            # NER tags are usually given per word not per word pieces
            and TokenClassification.__name__ not in self.head.config["type"]
            and RelationClassification.__name__ not in self.head.config["type"]
        ):
            self.tokenizer_config.use_transformers = True
            self.tokenizer_config.transformers_kwargs = (
                self.tokenizer_config.transformers_kwargs
                or {"model_name": self.features.transformers.model_name}
            )
        else:
            self.tokenizer_config.use_transformers = False

    def _check_for_incompatible_configurations(self):
        if self.tokenizer_config.use_transformers:
            if self.features.word is not None or self.features.char is not None:
                raise ConfigurationError(
                    "You are trying to use word or char features on subwords and possibly special tokens."
                    "This is not recommended!"
                )

            if (
                TokenClassification.__name__ in self.head.config["type"]
                or RelationClassification.__name__ in self.head.config["type"]
            ):
                raise NotImplementedError(
                    "You specified a transformers tokenizer, but the 'TokenClassification' and "
                    "'RelationClassification' heads are still not capable of dealing with subword/special tokens."
                )

            if (
                self.tokenizer_config.transformers_kwargs["model_name"]
                != self.features.transformers.model_name
            ):
                raise ConfigurationError(
                    f"The model_name of the TransformerTokenizer "
                    f"({self.tokenizer_config.transformers_kwargs['model_name']}) does not match the model_name of "
                    f"your transformers feature ({self.features.transformers.model_name})!"
                )

    @classmethod
    def from_yaml(cls, path: str) -> "PipelineConfiguration":
        """Creates a pipeline configuration from a config yaml file

        Parameters
        ----------
        path
            The path to a YAML configuration file

        Returns
        -------
        pipeline_configuration
        """
        with open(path) as yaml_file:
            config_dict = yaml.safe_load(yaml_file)

        return cls.from_dict(config_dict)

    @classmethod
    def from_dict(cls, config_dict: dict) -> "PipelineConfiguration":
        """Creates a pipeline configuration from a config dictionary

        Parameters
        ----------
        config_dict
            A configuration dictionary

        Returns
        -------
        pipeline_configuration
        """
        config_dict = sanitize_for_params(copy.deepcopy(config_dict))
        return PipelineConfiguration.from_params(Params(config_dict))

    def as_dict(self) -> Dict[str, Any]:
        """Returns the configuration as dictionary

        Returns
        -------
        config
        """
        config = {
            "name": self.name,
            "tokenizer": vars(self.tokenizer_config),
            "features": vars(self.features),
            "head": self.head.config,
        }

        if self.encoder:
            config["encoder"] = self.encoder.config

        return sanitize_for_params(config)

    def to_yaml(self, path: str):
        """Saves the pipeline configuration to a yaml formatted file

        Parameters
        ----------
        path
            Path to the output file
        """
        config_dict = copy.deepcopy(self.as_dict())
        for feature_name, feature in config_dict["features"]:
            if feature is not None:
                config_dict["features"][feature_name] = feature.to_json()

        save_dict_as_yaml(config_dict, path)

    def build_tokenizer(self) -> Tokenizer:
        """Build the pipeline tokenizer"""
        if self.tokenizer_config.use_transformers:
            return TransformersTokenizer(self.tokenizer_config)
        return Tokenizer(self.tokenizer_config)

    def build_featurizer(self) -> InputFeaturizer:
        """Creates the pipeline featurizer"""
        return self.features.compile_featurizer(self.build_tokenizer())

    def build_embedder(self, vocab: Vocabulary):
        """Build the pipeline embedder for aiven dictionary"""
        return self.features.compile_embedder(vocab)


@dataclass
class VocabularyConfiguration:
    """Configurations for creating the vocabulary

    See [AllenNLP Vocabulary docs](https://docs.allennlp.org/master/api/data/vocabulary/#vocabulary])

    Parameters
    ----------
    include_valid_data
        If passed to the `Trainer`, this argument allows you to take the validation data into account when creating
        the vocabulary (apart from the training data). Default: False.
    max_vocab_size
        If you want to cap the number of tokens in your vocabulary, you can do so with this
        parameter.  If you specify a single integer, every namespace will have its vocabulary fixed
        to be no larger than this.  If you specify a dictionary, then each namespace in the
        `counter` can have a separate maximum vocabulary size. Any missing key will have a value
        of `None`, which means no cap on the vocabulary size.
    min_count
        Minimum number of appearances of a token to be included in the vocabulary.
        The key in the dictionary refers to the namespace of the input feature
    min_pretrained_embeddings
        Minimum number of lines to keep from pretrained_files, even for tokens not appearing in the sources.
    only_include_pretrained_words
        Only include tokens present in pretrained_files
    pretrained_files
        If provided, this map specifies the path to optional pretrained embedding files for each
        namespace. This can be used to either restrict the vocabulary to only words which appear
        in this file, or to ensure that any words in this file are included in the vocabulary
        regardless of their count, depending on the value of `only_include_pretrained_words`.
        Words which appear in the pretrained embedding file but not in the data are NOT included
        in the Vocabulary.
    tokens_to_add
        A list of tokens to add to the corresponding namespace of the vocabulary,
        even if they are not present in the `datasets`
    """

    include_valid_data: bool = False
    max_vocab_size: Union[int, Dict[str, int]] = None
    min_count: Dict[str, int] = None
    min_pretrained_embeddings: Dict[str, int] = None
    only_include_pretrained_words: bool = False
    pretrained_files: Optional[Dict[str, str]] = None
    tokens_to_add: Dict[str, List[str]] = None


# We need this to be hashable for the prediction cache -> frozen=True
@dataclass(frozen=True)
class PredictionConfiguration:
    """Contains configurations for a `Pipeline.prediction`

    Parameters
    ----------
    add_tokens
    add_attributions
    attributions_kwargs
    """

    add_tokens: bool = False
    add_attributions: bool = False
    attributions_kwargs: Dict = field(default_factory=dict)


@dataclass
class TrainerConfiguration:
    """Configuration for the `biome.text.Trainer`.

    The docs are mainly a copy from the
    [Lightning Trainer API](https://pytorch-lightning.readthedocs.io/en/stable/api/pytorch_lightning.trainer.trainer.html#pytorch_lightning.trainer.trainer.Trainer)
    with some additional parameters added.

    Parameters
    ----------
    accumulate_grad_batches
        Accumulates grads every k batches or as set up in the dict.

    add_early_stopping
        Adds a default `EarlyStopping` callback. To configure it, see `patience`, `monitor` and `monitor_mode`.
        Default: True

    add_csv_logger
        Adds a default CSV logger if `logger` is not False. Default: True

    add_lr_monitor
        Adds a default `LearningRateMonitor(logging_interval="step")` to the callbacks.
        By default (None), we will set this to true if you use either `warmup_steps` and/or `lr_decay`.

    add_tensorboard_logger
        Adds a default Tensorboard logger if `logger` is not False. Default: True

    add_wandb_logger
        Adds a default WandB logger if `logger` is not False and wandb is installed. Default: True

    auto_lr_find
        If set to True, will make trainer.tune() run a learning rate finder,
        trying to optimize initial learning for faster convergence. trainer.tune() method will
        set the suggested learning rate in self.lr or self.learning_rate in the LightningModule.
        To use a different key set a string instead of True with the key name.

    auto_scale_batch_size
        If set to True, will `initially` run a batch size
        finder trying to find the largest batch size that fits into memory.
        The result will be stored in self.batch_size in the LightningModule.
        Additionally, can be set to either `power` that estimates the batch size through
        a power search or `binsearch` that estimates the batch size through a binary search.

    batch_size
        Size of the batch.

    callbacks
        Add a callback or list of callbacks.

    checkpoint_callback
        Adds a default `ModelCheckpointWithVocab` callback if there is no user-defined ModelCheckpoint in
        `callbacks`. To configure it, see `save_top_k_checkpoints`, `monitor` and `monitor_mode`.
        Default: True.

    check_val_every_n_epoch
        Check val every n train epochs.

    default_root_dir
        Default path for logs and weights when no logger/ckpt_callback passed.
        Can be remote file paths such as 's3://mybucket/path' or 'hdfs://path/'
        Default: './training_logs'.

    fast_dev_run
        runs n if set to ``n`` (int) else 1 if set to ``True`` batch(es)
        of train, val and test to find any bugs (ie: a sort of unit test).

    flush_logs_every_n_steps
        How often to flush logs to disk (defaults to every 100 steps).

    gpus
        number of gpus to train on (int) or which GPUs to train on (list or str) applied per node.
        If None (default), we will use one GPU if available.

    gradient_clip_val
        0 means don't clip.

    limit_train_batches
        How much of training dataset to check (floats = percent, int = num_batches)

    limit_val_batches
        How much of validation dataset to check (floats = percent, int = num_batches)

    limit_test_batches
        How much of test dataset to check (floats = percent, int = num_batches)

    log_every_n_steps
        How often to log within steps (defaults to every 50 steps).

    logger
        Logger (or iterable collection of loggers) for experiment tracking.
        If not False, we will add some loggers by default, see `add_[csv, tensorboard, wandb]_logger`.
        Default: True

    lr_decay
        Either 'linear' or 'cosine'. After an optional warmup, decay the learning rate in the specified manner.
        Default None

        You can use more sophisticated schedulers by first instantiating the Trainer and then manually specifying a
        scheduler, for example:
        ```python
        from torch.optim.lr_scheduler import ReduceLROnPlateau
        pipeline = Pipeline.from_config(...)
        trainer = Trainer(...)
        pipeline.model.lr_scheduler = {
            "scheduler": ReduceLROnPlateau(pipeline.model.optimizer, ...),
            "monitor": "validation_loss",
            "strict": True,
        }
        trainer.fit()
        ```
        See also https://pytorch-lightning.readthedocs.io/en/stable/common/optimizers.html#learning-rate-scheduling
        If specifying a scheduler manually, `lr_decay` and `warmup_steps` will have no effect.

    max_epochs
        Stop training once this number of epochs is reached. Disabled by default (None).
        If both max_epochs and max_steps are not specified, defaults to ``max_epochs`` = 1000.

    min_epochs
        Force training for at least these many epochs. Disabled by default (None).
        If both min_epochs and min_steps are not specified, defaults to ``min_epochs`` = 1.

    max_steps
        Stop training after this number of steps. Disabled by default (None).

    min_steps
        Force training for at least these number of steps. Disabled by default (None)

    monitor
        Metric to monitor. Will be used to load the best weights after the training (`checkpoint_callback` must be True)
        or stop the training early (`add_early_stopping` must be True). Default: 'validation_loss'.

    monitor_mode
        Either 'min' or 'max'. If `save_top_k_checkpoints != 0`, the decision to overwrite the current save file is made
        based on either the maximization or the minimization of the monitored metric (`checkpoint_callback` must be
        True). It also configures the default early stopping callback (`add_early_stopping` must be True).
        Default: 'min'

    num_workers_for_dataloader
        How many subprocesses to use for data loading. 0 means that the data will be loaded in the main process.
        Default: 0

    num_sanity_val_steps
        Sanity check runs n validation batches before starting the training routine.
        Set it to `-1` to run all batches in all validation dataloaders. Default: 2

    optimizer
        Configuration for an [AllenNLP/PyTorch optimizer](https://docs.allennlp.org/main/api/training/optimizers/)
        that is constructed via the AllenNLP configuration framework.
        Default: `{"type": "adam", "lr": 0.001}`

    overfit_batches
        Overfit a percent of training data (float) or a set number of batches (int). Default: 0.0

    patience
        Number of validation checks with no improvement after which training will be stopped. Default: 3.

    precision
        Full precision (32), half precision (16). Can be used on CPU, GPU or TPUs.

    progress_bar_refresh_rate
        How often to refresh progress bar (in steps). Value ``0`` disables progress bar.
        Ignored when a custom progress bar is passed to :paramref:`~Trainer.callbacks`. Default: None, means
        a suitable value will be chosen based on the environment (terminal, Google COLAB, etc.).

    resume_from_checkpoint
        Path/URL of the checkpoint from which training is resumed. If there is
        no checkpoint file at the path, start from scratch. If resuming from mid-epoch checkpoint,
        training will start from the beginning of the next epoch.

    save_top_k_checkpoints
        If `save_top_k_checkpoints == k`, the best k models according to the metric monitored will be saved.
        If `save_top_k_checkpoints == 0`, no models are saved. If `save_top_k_checkpoints == -1`, all models are saved.
        Has no effect if `checkpoint_callback` is False. Default: 1.

    stochastic_weight_avg
        Whether to use `Stochastic Weight Averaging (SWA)
        <https://pytorch.org/blog/pytorch-1.6-now-includes-stochastic-weight-averaging/>_`

    terminate_on_nan
        If set to True, will terminate training (by raising a `ValueError`) at the
        end of each training batch, if any of the parameters or the loss are NaN or +/-inf.

    val_check_interval
        How often to check the validation set. Use float to check within a training epoch,
        use int to check every n steps (batches).

    warmup_steps
        Number of steps for the warmup phase. In this initial phase the learning rate will be increased linearly from
        zero until the learning rate specified in the optimizer. See also `lr_decay` for more sophisticated schedulers.
        Default: 0

    weights_save_path
        Where to save weights if specified. Will override default_root_dir
        for checkpoints only. Use this if for whatever reason you need the checkpoints
        stored in a different place than the logs written in `default_root_dir`.
        Can be remote file paths such as `s3://mybucket/path` or 'hdfs://path/'
        Defaults to `default_root_dir`.

    extra_lightning_params
        This dictionary is passed on as kwargs to the Pytorch Lightning Trainer init method.
    """

    accumulate_grad_batches: Union[int, Dict[int, int], List[list]] = 1
    auto_lr_find: Union[bool, str] = False
    auto_scale_batch_size: Union[str, bool] = False
    callbacks: Optional[Union[List[Callback], Callback]] = None
    check_val_every_n_epoch: int = 1
    checkpoint_callback: bool = True
    default_root_dir: str = field(
        default_factory=lambda: os.path.join(os.getcwd(), "training_logs")
    )
    fast_dev_run: Union[int, bool] = False
    flush_logs_every_n_steps: int = 100
    gpus: Optional[Union[List[int], str, int]] = None
    gradient_clip_val: float = 0
    limit_train_batches: Union[int, float] = 1.0
    limit_val_batches: Union[int, float] = 1.0
    limit_test_batches: Union[int, float] = 1.0
    log_every_n_steps: int = 50
    logger: Union[LightningLoggerBase, Iterable[LightningLoggerBase], bool] = True
    max_epochs: Optional[int] = None
    min_epochs: Optional[int] = None
    max_steps: Optional[int] = None
    min_steps: Optional[int] = None
    num_sanity_val_steps: int = 2
    overfit_batches: Union[int, float] = 0.0
    precision: int = 32
    progress_bar_refresh_rate: Optional[int] = None
    resume_from_checkpoint: Optional[Union[Path, str]] = None
    stochastic_weight_avg: bool = False
    terminate_on_nan: bool = False
    val_check_interval: Union[int, float] = 1.0
    weights_save_path: Optional[str] = None
    # non lightning trainer parameters
    add_early_stopping: bool = True
    add_csv_logger: bool = True
    add_lr_monitor: Optional[bool] = None
    add_tensorboard_logger: bool = True
    add_wandb_logger: bool = True
    batch_size: int = 16
    lr_decay: Optional[str] = None
    monitor: str = "validation_loss"
    monitor_mode: str = "min"
    num_workers_for_dataloader: int = 0
    optimizer: Dict[str, Any] = field(
        default_factory=lambda: {"type": "adam", "lr": 0.001}
    )
    patience: int = 3
    save_top_k_checkpoints: int = 1
    warmup_steps: int = 0
    extra_lightning_params: Dict[str, Any] = field(default_factory=dict)

    def as_dict(self) -> Dict:
        """Returns the dataclass as dict without a deepcopy, in contrast to `dataclasses.asdict`"""
        return {fld.name: getattr(self, fld.name) for fld in fields(self)}

    def to_yaml(self, path: Union[str, Path]):
        """Saves the configuration as a yaml file"""
        if isinstance(path, str):
            path = Path(path)
        with path.open("w") as yml_file:
            yaml.dump(
                sanitize_for_yaml(self.as_dict()),
                yml_file,
                default_flow_style=False,
                allow_unicode=True,
            )

    @property
    def lightning_params(self) -> Dict[str, Any]:
        non_lightning_params = [
            "add_early_stopping",
            "add_csv_logger",
            "add_lr_monitor",
            "add_tensorboard_logger",
            "add_wandb_logger",
            "batch_size",
            "lr_decay",
            "monitor",
            "monitor_mode",
            "num_workers_for_dataloader",
            "optimizer",
            "patience",
            "save_top_k_checkpoints",
            "warmup_steps",
            "extra_lightning_params",
        ]

        lightning_params = {
            key: value
            for key, value in self.as_dict().items()
            if key not in non_lightning_params
        }
        lightning_params.update(self.extra_lightning_params)

        return lightning_params



from pathlib import Path

import mlflow
import pandas as pd
import pytest
import yaml
from numpy.testing import assert_allclose

from biome.text import Pipeline
from biome.text import __version__


@pytest.fixture
def pipeline():
    return Pipeline.from_config(
        {
            "name": "test_pipeline_copy",
            "head": {"type": "TextClassification", "labels": ["a", "b"]},
        }
    )


def test_to_mlflow(pipeline, tmp_path):
    test_str_for_prediction = "test this prediction"
    expected_prediction = pipeline.predict(text=test_str_for_prediction)

    model_uri = pipeline.to_mlflow(
        tracking_uri=str(tmp_path / "to_mlflow_test"), experiment_id=0
    )

    df = mlflow.search_runs(experiment_ids=["0"])
    assert len(df) == 1 and df["tags.mlflow.runName"][0] == "log_biometext_model"

    # load MLFlow model and make predictions
    model = mlflow.pyfunc.load_model(model_uri=model_uri)
    prediction: pd.DataFrame = model.predict(
        pd.DataFrame([{"text": test_str_for_prediction}])
    )

    assert len(prediction) == 1
    assert expected_prediction["labels"] == prediction["labels"][0]
    assert_allclose(
        expected_prediction["probabilities"], prediction["probabilities"][0]
    )
    with (Path(model_uri) / "conda.yaml").open() as file:
        conda_env = yaml.load(file)
        assert conda_env == {
            "name": "mlflow-dev",
            "channels": ["defaults", "conda-forge"],
            "dependencies": [
                "python=3.7.9",
                "pip>=20.3.0",
                {"pip": [f"biome-text=={__version__}"]},
            ],
        }



import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("biome-text").version
except pkg_resources.DistributionNotFound:
    # package is not installed
    pass

import logging

# configure basic 'biome.text' logging
_handler = logging.StreamHandler()
_handler.setFormatter(
    logging.Formatter("%(levelname)s:%(name)s: %(message)s")
)  # "%(levelname)s: %(message)s"))
_LOGGER = logging.getLogger(__name__)
_LOGGER.addHandler(_handler)
_LOGGER.setLevel("INFO")
# configure 'allennlp' logging
_ALLENNLP_LOGGER = logging.getLogger("allennlp")
_ALLENNLP_LOGGER.addHandler(_handler)
_ALLENNLP_LOGGER.setLevel("WARNING")

from .configuration import PipelineConfiguration
from .configuration import TrainerConfiguration
from .configuration import VocabularyConfiguration
from .dataset import Dataset
from .pipeline import Pipeline
from .trainer import Trainer



from pathlib import Path

import pandas as pd
import pytest
from allennlp.data import Instance
from datasets.features import Features
from datasets.features import Value

from biome.text import Dataset
from biome.text import Pipeline
from biome.text.dataset import AllennlpDataset
from biome.text.dataset import AllennlpLazyDataset


@pytest.fixture(scope="class")
def dataset(tmp_path_factory, request):
    cache_path = tmp_path_factory.mktemp("test_instance_caching")
    data_path = cache_path / "data.csv"
    data_path.write_text("text,label\ncheck,this\nand,that")

    dataset = Dataset.from_csv(paths=str(data_path), cache_dir=str(cache_path))

    # inject to classes decorated with `pytest.mark.usefixtures`
    if request.cls:
        request.cls.dataset = dataset

    return dataset


@pytest.fixture
def default_pipeline_config():
    return {
        "name": "datasets_test",
        "features": {"word": {"embedding_dim": 2}},
        "head": {"type": "TextClassification", "labels": ["this", "that"]},
    }


def test_load_dataset(resources_data_path):
    json_path = str(resources_data_path / "dataset_sequence.jsonl")

    with pytest.raises(TypeError):
        Dataset.load_dataset("json", data_files=[json_path])

    ds = Dataset.load_dataset("json", data_files=[json_path], split="train")
    assert len(ds) == 4


def test_from_json(resources_data_path):
    json_path = str(resources_data_path / "dataset_sequence.jsonl")
    ds = Dataset.from_json(paths=json_path)
    ds2 = Dataset.from_json(paths=[json_path, json_path])

    assert len(ds) == 4
    assert len(ds2) == 8

    json_path = str(resources_data_path / "dataset_sequence.json")
    ds = Dataset.from_json(paths=json_path, field="data")

    assert len(ds) == 4


def test_flatten_json(resources_data_path):
    """Showcases the behavior of Dataset.flatten_"""
    file_path = str(resources_data_path / "to-be-flattened.jsonl")
    dataset_flatten_source = Dataset.from_json(paths=file_path)
    dataset_flatten_source.flatten_()

    for c in ["complexData.a", "complexData.b"]:
        assert c in dataset_flatten_source.column_names

    file_path = str(resources_data_path / "nested-list.jsonl")
    dataset_nested_list = Dataset.from_json(paths=file_path)
    dataset_nested_list.flatten_()

    assert len(dataset_nested_list) == 1
    assert dataset_nested_list.column_names == ["classification"]


def test_from_csv(resources_data_path):
    csv_path = str(resources_data_path / "business.cat.2k.valid.csv")
    ds = Dataset.from_csv(paths=csv_path)
    ds2 = Dataset.from_csv(paths=[csv_path, csv_path])

    assert len(ds) == 400
    assert len(ds2) == 800
    assert ds.dataset.column_names == ["label", "text"]


def test_from_pandas():
    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    ds = Dataset.from_pandas(df)

    assert ds.dataset.column_names == ["a", "b"]
    assert ds["a"] == [1, 2, 3]
    assert len(ds) == 3


def test_from_dict():
    ds = Dataset.from_dict({"a": [1, 2, 3], "b": [4, 5, 6]})

    assert ds.dataset.column_names == ["a", "b"]
    assert ds["a"] == [1, 2, 3]
    assert len(ds) == 3


def test_from_excel_file(resources_data_path):
    """This only shows an example of how one could read in an excel file"""
    str_value = Value("string")
    int_value = Value("int64")
    features = Features(
        Notification=int_value, Type=str_value, Plant=int_value, Serial=str_value
    )

    file_path = resources_data_path / "test.xlsx"
    df = pd.read_excel(file_path)

    dataset = Dataset.from_pandas(df, features=features)

    assert len(dataset) > 0


def test_from_parquet_file(resources_data_path):
    """This only shows an example of how one could read in a parquet file"""
    file_path = resources_data_path / "test.parquet"
    df = pd.read_parquet(file_path)
    dataset = Dataset.from_pandas(df)

    assert "reviewerID" in dataset.column_names


def test_fail_using_reserved_words():
    # This issue was reported and maybe gets resolved: https://github.com/huggingface/datasets/issues/1110

    ds = Dataset.from_dict(
        {
            "a": [{"a": 1, "b": "two"} for _ in range(0, 100)],
            "_type": ["whatever" for _ in range(0, 100)],
        }
    )
    split = ds.train_test_split()
    new_ds = Dataset.from_datasets(split.values())

    assert len(new_ds) == len(ds)

    ds = ds.map(
        lambda example: {
            "new_field": {"c": str(example["a"]["a"]), "d": f"this {example['a']['b']}"}
        }
    )
    split = ds.train_test_split()
    with pytest.raises(TypeError):
        Dataset.from_datasets(split.values())


def test_to_instances(dataset, default_pipeline_config):
    pl = Pipeline.from_config(default_pipeline_config)

    instances = dataset.to_instances(pl, lazy=True)
    assert isinstance(instances, AllennlpLazyDataset)

    instances = dataset.to_instances(pl)
    assert isinstance(instances, AllennlpDataset)

    assert len(instances) == len(dataset)
    for instance in instances:
        assert isinstance(instance, Instance)
        assert "text" in instance.fields
        assert "label" in instance.fields

    # TODO: A missing label column should actually raise something ... but not so easy to implement
    dataset.rename_column_("label", "not_label")
    dataset.to_instances(pl)

    # Missing input column should raise a KeyError
    dataset.rename_column_("text", "not_text")
    with pytest.raises(KeyError):
        dataset.to_instances(pl)


@pytest.mark.usefixtures("dataset")
class TestInstanceCaching:
    def uses_cached_instances(self, pipeline_config) -> bool:
        """Checks if the `to_instances` method of the provided pipeline_config uses the cached instances"""
        cache_path = Path(self.dataset.dataset.cache_files[0]["filename"]).parent

        number_of_files_before = len(list(cache_path.iterdir()))
        pipeline = Pipeline.from_config(pipeline_config)
        self.dataset.to_instances(pipeline)
        number_of_files_after = len(list(cache_path.iterdir()))

        return number_of_files_before == number_of_files_after

    def test_same_pipeline(self, default_pipeline_config):
        self.dataset.cleanup_cache_files()
        assert not self.uses_cached_instances(default_pipeline_config)

        if not self.uses_cached_instances(default_pipeline_config):
            pytest.fail("The same pipelines did not reuse the cached instances!")

    def test_compatible_pipelines(self, default_pipeline_config):
        self.dataset.cleanup_cache_files()
        assert not self.uses_cached_instances(default_pipeline_config)

        default_pipeline_config["features"]["word"]["embedding_dim"] = 4
        if not self.uses_cached_instances(default_pipeline_config):
            pytest.fail("Compatible pipelines did not reuse the cached instances!")

    def test_incompatible_pipelines(self, default_pipeline_config):
        self.dataset.cleanup_cache_files()
        assert not self.uses_cached_instances(default_pipeline_config)

        default_pipeline_config["features"]["word"]["lowercase_tokens"] = True
        if self.uses_cached_instances(default_pipeline_config):
            pytest.fail("Incompatible pipelines did reuse the cached instances!")

        default_pipeline_config["features"]["char"] = {
            "embedding_dim": 2,
            "encoder": {"type": "gru", "hidden_size": 2},
        }
        if self.uses_cached_instances(default_pipeline_config):
            pytest.fail("Incompatible pipelines did reuse the cached instances!")

        default_pipeline_config["tokenizer"] = {"max_sequence_length": 10}
        if self.uses_cached_instances(default_pipeline_config):
            pytest.fail("Incompatible pipelines did reuse the cached instances!")

    def test_incompatible_datasets(self, default_pipeline_config):
        self.dataset.cleanup_cache_files()
        assert not self.uses_cached_instances(default_pipeline_config)

        self.dataset = self.dataset.map(
            lambda x: {}
        )  # this generates a new fingerprint for the dataset
        if self.uses_cached_instances(default_pipeline_config):
            pytest.fail("Incompatible datasets did reuse the cached instances!")

    def test_incompatible_versions(self, default_pipeline_config, monkeypatch):
        from biome.text import dataset

        self.dataset.cleanup_cache_files()
        assert not self.uses_cached_instances(default_pipeline_config)

        monkeypatch.setattr(dataset, "biome__version__", "mockversion")
        if self.uses_cached_instances(default_pipeline_config):
            pytest.fail(
                "Incompatible biome.text versions did reuse the cached instances!"
            )

        monkeypatch.setattr(dataset, "allennlp__version__", "mockversion")
        if self.uses_cached_instances(default_pipeline_config):
            pytest.fail(
                "Incompatible allennlp versions did reuse the cached instances!"
            )

        monkeypatch.setattr(dataset, "spacy__version__", "mockversion")
        if self.uses_cached_instances(default_pipeline_config):
            pytest.fail("Incompatible spacy versions did reuse the cached instances!")



import mlflow
import pytest
from ray import tune

from biome.text import Pipeline
from biome.text import TrainerConfiguration
from biome.text.dataset import Dataset
from biome.text.hpo import TuneExperiment


@pytest.fixture
def dataset():
    return Dataset.from_dict({"text": ["a", "b"], "label": ["a", "b"]})


@pytest.fixture
def pipeline_config():
    return {
        "name": "test_ray_tune_trainable",
        "features": {
            "word": {"embedding_dim": 2},
        },
        "head": {"type": "TextClassification", "labels": ["a", "b"]},
    }


@pytest.fixture
def trainer_config() -> TrainerConfiguration:
    return TrainerConfiguration(
        optimizer={"type": "adam", "lr": 0.01}, max_epochs=1, batch_size=2
    )


def test_tune_exp_default_trainable(
    dataset, pipeline_config, trainer_config, monkeypatch
):
    pipeline_config["features"]["word"]["embedding_dim"] = tune.choice([2, 4])
    trainer_config.optimizer["lr"] = tune.loguniform(0.001, 0.01)

    my_exp = TuneExperiment(
        pipeline_config=pipeline_config,
        trainer_config=trainer_config,
        train_dataset=dataset,
        valid_dataset=dataset,
        num_samples=1,
    )

    assert my_exp._name.startswith("HPO on")
    assert my_exp.name == my_exp._name
    assert my_exp._run_identifier == "_default_trainable"

    analysis = tune.run(my_exp)
    assert len(analysis.trials) == 1


def test_tune_exp_save_dataset_and_vocab(
    dataset, pipeline_config, trainer_config, monkeypatch
):
    pl = Pipeline.from_config(pipeline_config)
    vocab = pl.create_vocab([dataset.to_instances(pl)])

    my_exp = TuneExperiment(
        pipeline_config=pipeline_config,
        trainer_config=trainer_config,
        train_dataset=dataset,
        valid_dataset=dataset,
        vocab=vocab,
    )

    config = my_exp.config
    pl2 = Pipeline.from_config(config["pipeline_config"], config["vocab_path"])

    pl._model.extend_vocabulary(vocab)
    assert pl.backbone.vocab._index_to_token == pl2.backbone.vocab._index_to_token
    assert pl.backbone.vocab._token_to_index == pl2.backbone.vocab._token_to_index

    assert dataset[:] == Dataset.load_from_disk(config["train_dataset_path"])[:]
    assert dataset[:] == Dataset.load_from_disk(config["valid_dataset_path"])[:]


def test_tune_exp_custom_trainable(
    dataset,
    pipeline_config,
    trainer_config,
):
    def my_trainable(config):
        pass

    my_exp = TuneExperiment(
        pipeline_config=pipeline_config,
        trainer_config=trainer_config,
        train_dataset=dataset,
        valid_dataset=dataset,
        name="custom trainable",
        trainable=my_trainable,
    )

    assert my_exp.name == "custom trainable"
    assert my_exp.trainable == my_trainable
    assert my_exp._run_identifier == "my_trainable"



"""
    Manages vocabulary tasks and fetches vocabulary information

    Provides utilities for getting information from a given vocabulary.

    Provides management actions such as extending the labels, setting new labels or creating an "empty" vocab.
"""
import logging
from typing import Dict
from typing import List

from allennlp.data import Vocabulary
from allennlp.data.vocabulary import DEFAULT_NON_PADDED_NAMESPACES

from biome.text.features import TransformersFeatures
from biome.text.features import WordFeatures

LABELS_NAMESPACE = "gold_labels"

_LOGGER = logging.getLogger(__name__)


def get_labels(vocab: Vocabulary) -> List[str]:
    """Gets list of labels in the vocabulary

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`

    Returns
    -------
    labels: `List[str]`
        A list of label strings
    """
    return [k for k in vocab.get_token_to_index_vocabulary(namespace=LABELS_NAMESPACE)]


def label_for_index(vocab: Vocabulary, idx: int) -> str:
    """Gets label string for a label `int` id

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`
    idx: `int
        the token index

    Returns
    -------
    label: `str`
        The string for a label id
    """
    return vocab.get_token_from_index(idx, namespace=LABELS_NAMESPACE)


def index_for_label(vocab: Vocabulary, label: str) -> int:
    """Gets the label `int` id for label string

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary``
    label: `str`
        the label

    Returns
    -------
    label_idx: `int`
        The label id for label string
    """
    return vocab.get_token_index(label, namespace=LABELS_NAMESPACE)


def get_index_to_labels_dictionary(vocab: Vocabulary) -> Dict[int, str]:
    """Gets a dictionary for turning label `int` ids into label strings

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`

    Returns
    -------
    labels: `Dict[int, str]`
        A dictionary to get fetch label strings from ids
    """
    return vocab.get_index_to_token_vocabulary(LABELS_NAMESPACE)


def words_vocab_size(vocab: Vocabulary) -> int:
    """Fetches the vocabulary size for the `words` namespace

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`

    Returns
    -------
    size: `int`
        The vocabulary size for the words namespace
    """
    return vocab.get_vocab_size(WordFeatures.namespace)


def extend_labels(vocab: Vocabulary, labels: List[str]):
    """Adds a list of label strings to the vocabulary

    Use this to add new labels to your vocabulary (e.g., useful for reusing the weights of an existing classifier)

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`
    labels: `List[str]`
        A list of strings containing the labels to add to an existing vocabulary
    """
    vocab.add_tokens_to_namespace(labels, namespace=LABELS_NAMESPACE)


def set_labels(vocab: Vocabulary, new_labels: List[str]):
    """Resets the labels in the vocabulary with a given labels string list

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`
    new_labels: `List[str]`
        The label strings to add to the vocabulary
    """
    for namespace_vocab in [
        vocab.get_token_to_index_vocabulary(LABELS_NAMESPACE),
        vocab.get_index_to_token_vocabulary(LABELS_NAMESPACE),
    ]:
        tokens = list(namespace_vocab.keys())
        for token in tokens:
            del namespace_vocab[token]

    extend_labels(vocab, new_labels)


def create_empty_vocabulary() -> Vocabulary:
    """Creates an empty Vocabulary with configured namespaces

    Returns
    -------
    empty_vocab
        The transformers namespace is added to the `non_padded_namespace`.
    """
    # Following is a hack, because AllenNLP handles the Transformers vocab differently!
    # The transformer vocab has its own padding and oov token, so we add it to the non_padded_namespaces.
    # AllenNLP gives its "transformer vocab" by default the "tags" namespace, which is a non_padded_namespace ...
    # If we do not do this, then writing the vocab to a file and loading it will fail, since AllenNLP will
    # look for its default OVV token in the vocab unless it is flagged as non_padded_namespace.
    # (see the doc string of `allennlp.data.token_indexers.PretrainedTransformerIndexer`)
    return Vocabulary(
        non_padded_namespaces=DEFAULT_NON_PADDED_NAMESPACES
        + (TransformersFeatures.namespace,)
    )


def is_empty(vocab: Vocabulary, namespaces: List[str]) -> bool:
    """Checks if at least one of the given namespaces has an empty vocab.

    Parameters
    ----------
    vocab
        The vocabulary
    namespaces
        Namespaces to check in the vocabulary

    Returns
    -------
    True if one or more namespaces have an empty vocab
    """
    # If a namespace does not exist in the vocab, a default one is created on the fly with a padding and oov token
    # We must drop the padding and out of vocab (oov) tokens -> 2 tokens
    return any([vocab.get_vocab_size(namespace) < 3 for namespace in namespaces])



from typing import Any
from typing import List
from typing import Optional
from typing import Union

from astroid import Instance

from biome.text import Pipeline
from biome.text import PipelineConfiguration
from biome.text.configuration import FeaturesConfiguration
from biome.text.modules.heads import TaskHeadConfiguration
from biome.text.modules.heads import TextClassification


class MyCustomHead(TextClassification):
    """Just a head renaming the original TextClassification head"""

    def inputs(self) -> Optional[List[str]]:
        return ["text", "second_text"]

    def featurize(
        self,
        text: Any,
        second_text: Optional[Any] = None,
        label: Optional[Union[int, str, List[Union[int, str]]]] = None,
    ) -> Optional[Instance]:
        instance = self.backbone.featurizer(
            {"text": text, "text-2": second_text},
            to_field=self.forward_arg_name,
            aggregate=True,
            exclude_record_keys=True,
        )
        return self._add_label(instance, label, to_field=self.label_name)


def test_check_pipeline_inputs_and_output():
    config = PipelineConfiguration(
        "test-pipeline",
        head=TaskHeadConfiguration(
            type=MyCustomHead,
            labels=[
                "blue-collar",
                "technician",
                "management",
                "services",
                "retired",
                "admin.",
            ],
        ),
        features=FeaturesConfiguration(),
    )

    pipeline = Pipeline.from_config(config)

    assert pipeline.inputs == ["text", "second_text"]
    assert pipeline.output == ["label"]



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import numpy
import torch
from allennlp.data import Instance
from allennlp.data import TextFieldTensors
from allennlp.modules import FeedForward
from allennlp.modules import Seq2SeqEncoder
from allennlp.modules import Seq2VecEncoder
from allennlp.modules.bimpm_matching import BiMpmMatching
from allennlp.nn import InitializerApplicator
from allennlp.nn import util
from captum.attr import IntegratedGradients

from biome.text.backbone import ModelBackbone
from biome.text.modules.configuration import BiMpmMatchingConfiguration
from biome.text.modules.configuration import ComponentConfiguration
from biome.text.modules.configuration import FeedForwardConfiguration
from biome.text.modules.configuration import Seq2SeqEncoderConfiguration
from biome.text.modules.configuration import Seq2VecEncoderConfiguration
from biome.text.modules.encoders import TimeDistributedEncoder
from biome.text.modules.heads.classification.classification import ClassificationHead
from biome.text.modules.heads.task_prediction import Attribution
from biome.text.modules.heads.task_prediction import RecordPairClassificationPrediction


class RecordPairClassification(ClassificationHead):
    """Classifies the relation between a pair of records using a matching layer.

    The input for models using this `TaskHead` are two *records* with one or more *data fields* each, and a label
    describing their relationship.
    If you would like a meaningful explanation of the model's prediction,
    both records must consist of the same number of *data fields* and hold them in the same order.

    The architecture is loosely based on the AllenNLP implementation of the BiMPM model described in
    `Bilateral Multi-Perspective Matching for Natural Language Sentences <https://arxiv.org/abs/1702.03814>`_
    by Zhiguo Wang et al., 2017, and was adapted to deal with record pairs.

    Parameters
    ----------
    backbone
        Takes care of the embedding and optionally of the language encoding
    labels
        A list of labels for your classification task.
    field_encoder
        Encodes a data field, contextualized within the field
    record_encoder
        Encodes data fields, contextualized within the record
    matcher_forward
        BiMPM matching for the forward output of the record encoder layer
    matcher_backward
        BiMPM matching for the backward output of the record encoder layer
    aggregator
        Aggregator of all BiMPM matching vectors
    classifier_feedforward
        Fully connected layers for classification.
        A linear output layer with the number of labels at the end will be added automatically!!!
    dropout
        Dropout percentage to use.
    initializer
        If provided, will be used to initialize the model parameters.
    label_weights
        A list of weights for each label. The weights must be in the same order as the `labels`.
        You can also provide a dictionary that maps the label to its weight. Default: None.
    """

    _RECORD1_ARG_NAME_IN_FORWARD = "record1"
    _RECORD2_ARG_NAME_IN_FORWARD = "record2"

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        field_encoder: Seq2VecEncoderConfiguration,
        record_encoder: Seq2SeqEncoderConfiguration,
        matcher_forward: BiMpmMatchingConfiguration,
        aggregator: Seq2VecEncoderConfiguration,
        classifier_feedforward: FeedForwardConfiguration,
        matcher_backward: BiMpmMatchingConfiguration = None,
        dropout: float = 0.1,
        initializer: InitializerApplicator = InitializerApplicator(),
        label_weights: Optional[Union[List[float], Dict[str, float]]] = None,
    ):
        super().__init__(backbone, labels, label_weights=label_weights)

        # This is needed for the TrainerConfig to choose the right 'sorting_keys'
        self.backbone.encoder = TimeDistributedEncoder(self.backbone.encoder)

        self._field_encoder: Seq2VecEncoder = field_encoder.input_dim(
            self.backbone.encoder.get_output_dim()
        ).compile()
        self._field_encoder = TimeDistributedEncoder(self._field_encoder)

        self._record_encoder: Seq2SeqEncoder = record_encoder.input_dim(
            self._field_encoder.get_output_dim()
        ).compile()

        input_dim_matcher = self._record_encoder.get_output_dim()
        if matcher_backward is not None:
            input_dim_matcher //= 2

        self._matcher_forward: BiMpmMatching = matcher_forward.input_dim(
            input_dim_matcher
        ).compile()
        self._matcher_backward = (
            matcher_backward.input_dim(input_dim_matcher).compile()
            if matcher_backward is not None
            else None
        )

        matching_dim = self._matcher_forward.get_output_dim()
        if self._matcher_backward:
            matching_dim += self._matcher_backward.get_output_dim()

        self._aggregator: Seq2VecEncoder = aggregator.input_dim(matching_dim).compile()

        # we aggregate the two records in the end -> *2
        self._classifier_feedforward: FeedForward = classifier_feedforward.input_dim(
            self._aggregator.get_output_dim() * 2
        ).compile()

        self._output_layer = torch.nn.Linear(
            self._classifier_feedforward.get_output_dim(), self.num_labels
        )

        self._dropout = torch.nn.Dropout(dropout)

        initializer(self)

    def featurize(
        self,
        record1: Dict[str, str],
        record2: Dict[str, str],
        label: Optional[str] = None,
    ) -> Instance:
        """Tokenizes, indexes and embeds the two records and optionally adds the label

        Parameters
        ----------
        record1
            First record
        record2
            Second record
        label
            Classification label

        Returns
        -------
        instance
            AllenNLP instance containing the two records plus optionally a label

        Raises
        ------
        FeaturizeError
        """
        record1_instance = self.backbone.featurizer(
            record1, to_field="record", aggregate=False
        )
        record2_instance = self.backbone.featurizer(
            record2, to_field="record", aggregate=False
        )

        instance = Instance(
            {
                self._RECORD1_ARG_NAME_IN_FORWARD: record1_instance.get("record"),
                self._RECORD2_ARG_NAME_IN_FORWARD: record2_instance.get("record"),
            }
        )

        return self._add_label(instance, label)

    def forward(
        self,  # type: ignore
        record1: TextFieldTensors,
        record2: TextFieldTensors,
        label: torch.LongTensor = None,
    ) -> Dict[str, Any]:
        # pylint: disable=arguments-differ
        """
        Parameters
        ----------
        record1
            Tokens of the first record.
            The dictionary is the output of a `ListField.as_array()`. It gives names to the tensors created by
            the `TokenIndexer`s.
            In its most basic form, using a `SingleIdTokenIndexer`, the dictionary is composed of:
            `{"tokens": {"tokens": Tensor(batch_size, num_fields, num_tokens)}}`.
            The dictionary is designed to be passed on directly to a `TextFieldEmbedder`, that has a
            `TokenEmbedder` for each key in the dictionary (except you set `allow_unmatched_keys` in the
            `TextFieldEmbedder` to False) and knows how to combine different word/character representations into a
            single vector per token in your input.
        record2
            Tokens of the second record.
        label : torch.LongTensor, optional (default = None)
            A torch tensor representing the sequence of integer gold class label of shape
            `(batch_size, num_classes)`.

        Returns
        -------
        An output dictionary consisting of:
        logits : torch.FloatTensor
        class_probabilities : torch.FloatTensor
        loss : torch.FloatTensor, optional
            A scalar loss to be optimised.
        """
        # embeding and encoding (field context)
        field_encoded_record1, record_mask_record1 = self._field_encoding(record1)
        field_encoded_record2, record_mask_record2 = self._field_encoding(record2)

        # encoding (record context), matching, aggregation, classification
        logits = self._bimpm_forward(
            field_encoded_record1,
            field_encoded_record2,
            record_mask_record1,
            record_mask_record2,
        )

        output = self._make_forward_output(logits, label)

        # For computing the attributions
        # TODO: An optimized implementation would be to calculate the attributions directly in the forward method
        #  and provide a practical switch, maybe: `with head.turn_attributions_on(): self.forward_on_instances()`
        #  In this way we would calculate the attributions batch wise and on on GPU if available.
        if not self.training:
            output["field_encoded_record1"] = field_encoded_record1
            output["record_mask_record1"] = record_mask_record1
            output["field_encoded_record2"] = field_encoded_record2
            output["record_mask_record2"] = record_mask_record2

        return output

    def _field_encoding(
        self,
        record: TextFieldTensors,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Embeds and encodes the records in a field context.

        We do this in a helper function to reuse it for the `self._compute_attributions` method.

        Parameters
        ----------
        record
            See `self.forward`

        Returns
        -------
        field_encoded_record
        record_mask
        """
        field_mask = util.get_text_field_mask(record, num_wrapping_dims=1)

        # embedding and encoding of one record (field context)
        embedded_record = self._dropout(
            self.backbone.forward(record, mask=field_mask, num_wrapping_dims=1)
        )
        field_encoded_record: torch.Tensor = self._dropout(
            self._field_encoder(embedded_record, mask=field_mask)
        )

        # mask for the record encoder
        record_mask = cast(torch.BoolTensor, torch.sum(field_mask, -1) > 0)

        return field_encoded_record, record_mask

    def _bimpm_forward(
        self,
        field_encoded_record1: torch.Tensor,
        field_encoded_record2: torch.Tensor,
        record_mask_record1: torch.Tensor,
        record_mask_record2: torch.Tensor,
    ) -> torch.Tensor:
        """Encodes in a record context, matches encodings, aggregates the matching results, classifies.

        We do this in a helper function to reuse it for the `self._compute_attributions` method.

        Parameters
        ----------
        field_encoded_record1
            Encoded record1 (in a field context)
        field_encoded_record2
            Encoded record2 (in a field context)
        record_mask_record1
            Mask for the record encoder for record1
        record_mask_record2
            Mask for the record encoder for record2

        Returns
        -------
        logits
        """
        record_encoded_record1 = self._dropout(
            self._record_encoder(field_encoded_record1, mask=record_mask_record1)
        )

        record_encoded_record2 = self._dropout(
            self._record_encoder(field_encoded_record2, mask=record_mask_record2)
        )

        # matching layer
        matching_vector_record1, matching_vector_record2 = self._matching_layer(
            record_encoded_record1,
            record_encoded_record2,
            record_mask_record1,
            record_mask_record2,
        )

        # aggregation layer
        aggregated_record1 = self._dropout(
            self._aggregator(matching_vector_record1, record_mask_record1)
        )
        aggregated_record2 = self._dropout(
            self._aggregator(matching_vector_record2, record_mask_record2)
        )

        aggregated_records = torch.cat([aggregated_record1, aggregated_record2], dim=-1)

        # the final feed forward layer
        logits = self._output_layer(self._classifier_feedforward(aggregated_records))

        return logits

    def _matching_layer(
        self,
        encoded_record1: torch.Tensor,
        encoded_record2: torch.Tensor,
        mask_record1: torch.Tensor,
        mask_record2: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Implements the Matching layer

        Parameters
        ----------
        encoded_record1
            encoded record1 (record context)
        encoded_record2
            encoded record2 (record context)
        mask_record1
            mask for encoded record1
        mask_record2
            mask for encoded record2

        Returns
        -------
        matching_vector_record1_cat
            Concatenated matching results for record1
        matching_vector_record2_cat
            Concatenated matching results for record2
        """
        matching_vector_record1: List[torch.Tensor] = []
        matching_vector_record2: List[torch.Tensor] = []

        # Matching is done separately for forward and backward encodings (if backward is specified)
        half_hidden_size = None
        if self._matcher_backward is not None:
            half_hidden_size = self._record_encoder.get_output_dim() // 2

        # forward matching
        matching_result = self._matcher_forward(
            encoded_record1[:, :, :half_hidden_size],
            mask_record1,
            encoded_record2[:, :, :half_hidden_size],
            mask_record2,
        )
        matching_vector_record1.extend(matching_result[0])
        matching_vector_record2.extend(matching_result[1])

        # backward matching
        if self._matcher_backward is not None:
            matching_result = self._matcher_backward(
                encoded_record1[:, :, half_hidden_size:],
                mask_record1,
                encoded_record2[:, :, half_hidden_size:],
                mask_record2,
            )
            matching_vector_record1.extend(matching_result[0])
            matching_vector_record2.extend(matching_result[1])

        # concat the matching vectors + apply dropout
        matching_vector_record1_cat = self._dropout(
            torch.cat(matching_vector_record1, dim=2)
        )
        matching_vector_record2_cat = self._dropout(
            torch.cat(matching_vector_record2, dim=2)
        )

        return matching_vector_record1_cat, matching_vector_record2_cat

    def _compute_attributions(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
        n_steps: int = 50,
    ) -> List[Attribution]:
        """Computes attributions for each data field in the record by means of the
        [Integrated Gradients](https://arxiv.org/abs/1703.01365) method.

        IMPORTANT: The calculated attributions only make sense for a duplicate/not_duplicate binary classification task
        of the two records.

        Parameters
        ----------
        single_forward_output
            Non-batched forward output containing numpy arrays
        instance
            The instance containing the input data
        n_steps
            The number of steps used when calculating the attribution of each token.

        Returns
        -------
        attributions
        """
        # captum needs `torch.Tensor`s and we need a batch dimension (-> unsqueeze)
        field_encoded_record1 = torch.from_numpy(
            single_forward_output["field_encoded_record1"]
        ).unsqueeze(0)
        record_mask_record1 = torch.from_numpy(
            single_forward_output["record_mask_record1"]
        ).unsqueeze(0)

        field_encoded_record2 = torch.from_numpy(
            single_forward_output["field_encoded_record2"]
        ).unsqueeze(0)
        record_mask_record2 = torch.from_numpy(
            single_forward_output["record_mask_record2"]
        ).unsqueeze(0)

        logits = torch.from_numpy(single_forward_output["logits"]).unsqueeze(0)

        if not field_encoded_record1.size() == field_encoded_record2.size():
            raise RuntimeError("Both records must have the same number of data fields!")

        # 2. Get attributes
        ig = IntegratedGradients(self._bimpm_forward)

        prediction_target = torch.argmax(logits)

        ig_attribute_record1 = ig.attribute(
            inputs=(field_encoded_record1, field_encoded_record2),
            baselines=(field_encoded_record2, field_encoded_record2),
            additional_forward_args=(record_mask_record1, record_mask_record2),
            target=prediction_target,
            return_convergence_delta=True,
            n_steps=n_steps,
        )

        ig_attribute_record2 = ig.attribute(
            inputs=(field_encoded_record1, field_encoded_record2),
            baselines=(field_encoded_record1, field_encoded_record1),
            additional_forward_args=(record_mask_record1, record_mask_record2),
            target=prediction_target,
            return_convergence_delta=True,
            n_steps=n_steps,
        )
        # The code below was an attempt to make attributions for the "duplicate case" more meaningful ... did not work
        # # duplicate case:
        # # Here we integrate each record along the path from the null vector -> record1/2
        # # assuming that the null vector provides the highest "not duplicate" score.
        # if prediction_target == 0:
        #     ig_attribute_record1 = ig.attribute(
        #         inputs=(field_encoded_record1, field_encoded_record2),
        #         baselines=(torch.zeros_like(field_encoded_record1), field_encoded_record2),
        #         additional_forward_args=(record_mask_record1, record_mask_record2),
        #         # we fix the target since we want negative integrals always to be associated
        #         # to the "not_duplicate" case and positive ones to the "duplicate" case
        #         target=0,
        #         return_convergence_delta=True,
        #     )
        #
        #     ig_attribute_record2 = ig.attribute(
        #         inputs=(field_encoded_record1, field_encoded_record2),
        #         baselines=(field_encoded_record1, torch.zeros_like(field_encoded_record2)),
        #         additional_forward_args=(record_mask_record1, record_mask_record2),
        #         # we fix the target since we want negative integrals always to be associated
        #         # to the "not_duplicate" case and positive ones to the "duplicate" case
        #         target=0,
        #         return_convergence_delta=True,
        #     )
        #
        # # not duplicate case:
        # # Here we integrate each record along the path from record2/1 -> record1/2
        # # assuming that the same record provides the highest "duplicate" score.
        # elif prediction_target == 1:
        #     ...
        # else:
        #     raise RuntimeError("The `compute_attributions` method is only implemented for a binary classification task: "
        #                        "[duplicate, not_duplicate]")

        attributions_record1, delta_record1 = self._get_attributions_and_delta(
            ig_attribute_record1, 0
        )
        attributions_record2, delta_record2 = self._get_attributions_and_delta(
            ig_attribute_record2, 1
        )

        # 3. Get tokens corresponding to the attributions
        field_text_record1 = []
        for textfield in instance.get(self._RECORD1_ARG_NAME_IN_FORWARD):
            field_text_record1.append(
                " ".join([token.text for token in textfield.tokens])
            )
        field_text_record2 = []
        for textfield in instance.get(self._RECORD2_ARG_NAME_IN_FORWARD):
            field_text_record2.append(
                " ".join([token.text for token in textfield.tokens])
            )

        output_record1 = [
            Attribution(
                text=field_text,
                start=0,
                end=len(field_text),
                field=self._RECORD1_ARG_NAME_IN_FORWARD,
                attribution=attribution,
            )
            for field_text, attribution in zip(field_text_record1, attributions_record1)
        ]
        output_record2 = [
            Attribution(
                text=field_text,
                start=0,
                end=len(field_text),
                field=self._RECORD2_ARG_NAME_IN_FORWARD,
                attribution=attribution,
            )
            for field_text, attribution in zip(field_text_record2, attributions_record2)
        ]

        return output_record1 + output_record2

    @staticmethod
    def _get_attributions_and_delta(
        ig_attribute_output, zero_or_one: int
    ) -> Tuple[numpy.array, float]:
        """Gets attributions and delta out of the `IntegratedGradients.attribute()` output.

        Parameters
        ----------
        ig_attribute_output
            Output of the `IntegratedGradients.attribute()` method.
        zero_or_one
            Is either 0 for record1 or 1 for record2

        Returns
        -------
        attributions, delta
        """
        attributions = ig_attribute_output[0][zero_or_one].sum(dim=2).squeeze(0)
        attributions = attributions / torch.norm(attributions)
        attributions = attributions.cpu().detach().numpy()
        delta = ig_attribute_output[1]

        return attributions, delta

    def _make_task_prediction(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
    ) -> RecordPairClassificationPrediction:
        labels, probabilities = self._compute_labels_and_probabilities(
            single_forward_output
        )

        return RecordPairClassificationPrediction(
            labels=labels, probabilities=probabilities
        )


class RecordPairClassificationConfiguration(
    ComponentConfiguration[RecordPairClassification]
):
    """Config for record pair classification head component"""

    pass



import pytest

from biome.text import Pipeline
from biome.text.configuration import PredictionConfiguration
from biome.text.model import PipelineModel
from biome.text.modules.heads.task_prediction import TaskPrediction


@pytest.fixture
def model() -> PipelineModel:
    pipeline = Pipeline.from_config(
        {
            "name": "test_predict",
            "head": {"type": "TextClassification", "labels": ["a"]},
        }
    )
    return pipeline._model


def test_activate_eval_mode(model):
    model.train()
    model.predict([{"text": "test"}], PredictionConfiguration)
    assert model.training is False


def test_forward_pass_error(model, monkeypatch, caplog):
    def mock_text_to_instance(**kwargs):
        return "mock instance"

    def mock_forward_on_instances(*args, **kwargs):
        raise Exception("mock Exception")

    monkeypatch.setattr(model, "text_to_instance", mock_text_to_instance)
    monkeypatch.setattr(model, "forward_on_instances", mock_forward_on_instances)

    predictions = model.predict(
        [{"text": "Some value that breaks the forward pass"}], PredictionConfiguration
    )

    assert predictions == [None]
    assert len(caplog.record_tuples) == 2
    assert caplog.record_tuples[0] == ("biome.text.model", 40, "mock Exception")
    assert caplog.record_tuples[1] == (
        "biome.text.model",
        30,
        "Failed to make a forward pass for '[{'text': 'Some value that breaks the forward pass'}]'",
    )


def test_return_type(model, monkeypatch):
    def mock_make_task_prediction(*args, **kwargs):
        return TaskPrediction()

    monkeypatch.setattr(model.head, "make_task_prediction", mock_make_task_prediction)

    predictions = model.predict(
        [{"text": "test"}, {"text": "test2"}], PredictionConfiguration()
    )
    assert isinstance(predictions, list)
    assert all([isinstance(pred, TaskPrediction) for pred in predictions])


def test_text_to_instance(model, caplog):
    with pytest.raises(TypeError):
        model.text_to_instance(wrong_kwarg="wrong argument")

    with pytest.raises(TypeError):
        model.text_to_instance(label="missing required argument")

    model.text_to_instance(text="")
    assert caplog.record_tuples[0] == (
        "biome.text.model",
        30,
        "The provided input data contains empty strings/tokens: ",
    )



from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import Optional

from biome.text.modules.configuration import Seq2VecEncoderConfiguration


class Features(ABC):
    """All features used in the pipeline configuration inherit from this abstract class."""

    @property
    @abstractmethod
    def config(self):
        pass

    @abstractmethod
    def to_json(self):
        pass

    def __eq__(self, other):
        return all([a == b for a, b in zip(vars(self), vars(other))])


class WordFeatures(Features):
    """Feature configuration at word level

    Parameters
    ----------
    embedding_dim
        Dimension of the embeddings
    lowercase_tokens
        If True, lowercase tokens before the indexing
    trainable
        If False, freeze the embeddings
    weights_file
        Path to a file with pretrained weights for the embedding
    **extra_params
        Extra parameters passed on to the `indexer` and `embedder` of the AllenNLP configuration framework.
        For example: `WordFeatures(embedding_dim=300, embedder={"padding_index": 0})`
    """

    namespace = "word"

    def __init__(
        self,
        embedding_dim: int,
        lowercase_tokens: bool = False,
        trainable: bool = True,
        weights_file: Optional[str] = None,
        **extra_params
    ):
        self.embedding_dim = embedding_dim
        self.lowercase_tokens = lowercase_tokens
        self.trainable = trainable
        self.weights_file = weights_file
        self.extra_params = extra_params

    @property
    def config(self) -> Dict:
        """Returns the config in AllenNLP format"""
        config = {
            "indexer": {
                "type": "single_id",
                "lowercase_tokens": self.lowercase_tokens,
                "namespace": self.namespace,
            },
            "embedder": {
                "embedding_dim": self.embedding_dim,
                "vocab_namespace": self.namespace,
                "trainable": self.trainable,
                "pretrained_file": self.weights_file,
            },
        }

        for k in self.extra_params:
            config[k] = {**self.extra_params[k], **config.get(k)}

        return config

    def to_json(self) -> Dict:
        """Returns the config as dict for the serialized json config file"""
        return {
            "embedding_dim": self.embedding_dim,
            "lowercase_tokens": self.lowercase_tokens,
            "trainable": self.trainable,
            "weights_file": self.weights_file,
            **self.extra_params,
        }


class CharFeatures(Features):
    """Feature configuration at character level

    Parameters
    ----------
    embedding_dim
        Dimension of the character embeddings.
    encoder
        A sequence to vector encoder resulting in a word representation based on its characters
    dropout
        Dropout applied to the output of the encoder
    lowercase_characters
        If True, lowercase characters before the indexing
    **extra_params
        Extra parameters passed on to the `indexer` and `embedder` of the AllenNLP configuration framework.
        For example: `CharFeatures(embedding_dim=32, indexer={"min_padding_length": 5}, ...)`
    """

    namespace = "char"

    def __init__(
        self,
        embedding_dim: int,
        encoder: Dict[str, Any],
        dropout: float = 0.0,
        lowercase_characters: bool = False,
        **extra_params
    ):
        self.embedding_dim = embedding_dim
        self.encoder = encoder
        self.dropout = dropout
        self.lowercase_characters = lowercase_characters
        self.extra_params = extra_params

    @property
    def config(self) -> Dict:
        """Returns the config in AllenNLP format"""
        if self.encoder["type"] == "cnn":
            min_padding_length = max(self.encoder.get("ngram_filter_sizes", [5]))
        elif self.encoder["type"] in ["gru", "lstm"]:
            min_padding_length = 1
        else:
            min_padding_length = 0

        config = {
            "indexer": {
                "type": "characters",
                "namespace": self.namespace,
                "character_tokenizer": {
                    "lowercase_characters": self.lowercase_characters
                },
                "min_padding_length": min_padding_length,
            },
            "embedder": {
                "type": "character_encoding",
                "embedding": {
                    "embedding_dim": self.embedding_dim,
                    "vocab_namespace": self.namespace,
                },
                "encoder": Seq2VecEncoderConfiguration(**self.encoder)
                .input_dim(self.embedding_dim)
                .config,
                "dropout": self.dropout,
            },
        }

        for k, v in self.extra_params.items():
            config[k] = {**self.extra_params[k], **config.get(k)}

        return config

    def to_json(self):
        """Returns the config as dict for the serialized json config file"""
        return {
            "embedding_dim": self.embedding_dim,
            "lowercase_characters": self.lowercase_characters,
            "encoder": self.encoder,
            "dropout": self.dropout,
            **self.extra_params,
        }


class TransformersFeatures(Features):
    """Configuration of the feature extracted with the [transformers models](https://huggingface.co/models).

    We use AllenNLPs "mismatched" indexer and embedder to get word-level representations.
    Most of the transformers models work with word-piece tokenizers.

    Parameters
    ----------
    model_name
        Name of one of the [transformers models](https://huggingface.co/models).
    trainable
        If false, freeze the transformer weights
    max_length
        If positive, split the document into segments of this many tokens (including special tokens)
        before feeding into the embedder. The embedder embeds these segments independently and
        concatenate the results to get the original document representation.
    last_layer_only
        When `True`, only the final layer of the pretrained transformer is taken
        for the embeddings. But if set to `False`, a scalar mix of all of the layers
        is used.

    Attributes
    ----------
    mismatched: bool (default: True)
        If true, we will sum up the word-piece vectors in a word to pull out a single vector for each word.
        If you use a transformers tokenizer, this is automatically set to False by the `PipelineConfiguration`.
    """

    namespace = "transformers"

    def __init__(
        self,
        model_name: str,
        trainable: bool = False,
        max_length: Optional[int] = None,
        last_layer_only: bool = True,
    ):
        self.model_name = model_name
        self.trainable = trainable
        self.max_length = max_length
        self.last_layer_only = last_layer_only
        self.mismatched = True

    @property
    def config(self) -> Dict:
        """Returns the config in AllenNLP format"""
        config = {
            "indexer": {
                "type": "pretrained_transformer_mismatched"
                if self.mismatched
                else "pretrained_transformer",
                "model_name": self.model_name,
                "namespace": self.namespace,
                "max_length": self.max_length,
            },
            "embedder": {
                "type": "pretrained_transformer_mismatched"
                if self.mismatched
                else "pretrained_transformer",
                "model_name": self.model_name,
                "train_parameters": self.trainable,
                "max_length": self.max_length,
                "last_layer_only": self.last_layer_only,
            },
        }

        return config

    def to_json(self) -> Dict:
        """Returns the config as dict for the serialized json config file"""
        return {
            "model_name": self.model_name,
            "trainable": self.trainable,
            "max_length": self.max_length,
            "last_layer_only": self.last_layer_only,
        }



import inspect
import json
import logging
import os
import pickle
import warnings
from functools import lru_cache
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

import allennlp
import pytorch_lightning as pl
import torch
from allennlp.common import Params
from allennlp.common.util import sanitize
from allennlp.data import Instance
from allennlp.data import Vocabulary
from allennlp.models.archival import CONFIG_NAME

from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.configuration import PipelineConfiguration
from biome.text.configuration import PredictionConfiguration
from biome.text.featurizer import FeaturizeError
from biome.text.helpers import split_signature_params_by_predicate
from biome.text.modules.heads import TaskHead
from biome.text.modules.heads import TaskPrediction


class _HashDict(dict):
    """
    Hashable dict implementation.
    BE AWARE! Since dicts are mutable, the hash can change!
    """

    def __hash__(self):
        return pickle.dumps(self).__hash__()


class _HashList(list):
    """
    Hashable list implementation.
    BE AWARE! Since lists are mutable, the hash can change!
    """

    def __hash__(self):
        return pickle.dumps(self).__hash__()


class PipelineModel(allennlp.models.Model, pl.LightningModule):
    """
    This class represents pipeline model implementation for connect biome.text concepts with
    allennlp implementation details

    This class manages the head + backbone encoder, keeping the allennlnlp Model lifecycle. This class
    should be hidden to api users.

    Parameters
    ----------
    config
        Configuration of the pipeline
    vocab
        The vocabulary of the pipeline. If None, an empty vocabulary will be created (default).

    Attributes
    ----------
    name: str
        Name of the pipeline model
    head: TaskHead
        TaskHead of the pipeline model
    vocab: Vocabulary
        The vocabulary of the model, comes from allennlp.models.Model
    file_path: Optional[str]
        File path to a serialized version of this pipeline model
    inputs: List[str]
        The model inputs
    output: List[str]
        The model outputs (not prediction): Corresponding to the `TaskHead.featurize` optional arguments.
    """

    PREDICTION_FILE_NAME = "predictions.json"
    TRAINING_METRICS_PREFIX = "training"
    VALIDATION_METRICS_PREFIX = "validation"
    TEST_METRICS_PREFIX = "test"
    _LOGGER = logging.getLogger(__name__)

    def __init__(self, config: Dict, vocab: Optional[Vocabulary] = None):
        super().__init__(vocab=vocab or vocabulary.create_empty_vocabulary())

        # saves the config in the pl checkpoints
        self.save_hyperparameters("config")

        config = PipelineConfiguration.from_dict(config)
        tokenizer = config.build_tokenizer()
        featurizer = config.features.compile_featurizer(tokenizer)
        embedder = config.build_embedder(self.vocab)
        head = config.head.compile(
            backbone=ModelBackbone(
                self.vocab,
                featurizer=featurizer,
                embedder=embedder,
                encoder=config.encoder,
            )
        )

        self.name = config.name
        self._head = None
        self.set_head(head)

        self.file_path: Optional[str] = None

        self.optimizer: Optional[torch.optim.Optimizer] = None
        # The lr_scheduler dict follows the Lightning format:
        # https://pytorch-lightning.readthedocs.io/en/stable/common/optimizers.html#learning-rate-scheduling
        self.lr_scheduler: Optional[Dict] = None

        self.best_metrics: Optional[Dict[str, torch.Tensor]] = None
        # This is set by our trainer to figure out the best_metrics
        # what metric to monitor?
        self.monitor: Optional[str] = None
        # shall the metric increase ("max") or decrease ("min")?
        self.monitor_mode: Optional[str] = None

    def _update_head_related_attributes(self):
        """Updates the inputs/outputs and default mapping attributes, calculated from model head"""
        required, optional = split_signature_params_by_predicate(
            self._head.featurize, lambda p: p.default == inspect.Parameter.empty
        )
        self._inputs = self._head.inputs() or [p.name for p in required]
        self._output = [p.name for p in optional if p.name not in self._inputs] or [
            None
        ]

    @classmethod
    def from_params(
        cls: "PipelineModel",
        params: Params,
        vocab: Optional[Vocabulary] = None,
        **extras,
    ) -> "PipelineModel":
        """
        Load the model implementation from params. We build manually each component from config sections.

        The param keys matches exactly with keys in yaml configuration files

        Parameters
        ----------
        params
            The config key in these params is used to build the model components
        vocab
            The vocabulary for the model
        **extras
            Necessary for AllenNLP from_params machinery

        Returns
        -------
        pipeline_model
        """

        config = params.pop("config", keep_as_dict=True)
        return cls(config=config, vocab=vocab)

    @property
    def head(self) -> TaskHead:
        """Get the model head"""
        return self._head

    def set_head(self, head: TaskHead) -> None:
        """Set a head and update related model attributes"""
        self._head = head
        self._update_head_related_attributes()

    def forward(self, *args, **kwargs) -> Dict[str, torch.Tensor]:
        """The main forward method just wraps the head forward method"""
        return self._head(*args, **kwargs)

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        """Fetch metrics defined in head layer"""
        return self._head.get_metrics(reset)

    def text_to_instance(self, **inputs) -> Optional[Instance]:
        """Applies the head featurize method"""
        try:
            return self._head.featurize(**inputs)
        except FeaturizeError as error:
            # we cannot featurize the input (empty strings, etc.)
            self._LOGGER.warning(error)
        except TypeError as error:
            # probably wrong input arguments for the head
            raise TypeError(
                f"Please check your input arguments, expected: {self.inputs}, actual: {inputs.keys()}"
            ) from error

    def extend_vocabulary(self, vocab: Vocabulary):
        """Extend the model's vocabulary with `vocab`

        Parameters
        ----------
        vocab
            The model's vocabulary will be extended with this one.
        """
        # self.vocab and self._head.backbone.vocab point to the same vocab!
        self.vocab.extend_from_vocab(vocab)

        # updates the embedding matrices
        self.extend_embedder_vocab()
        # updates head specific things
        self._head.on_vocab_update()

    def extend_embedder_vocab(
        self, embedding_sources_mapping: Dict[str, str] = None
    ) -> None:
        """
        Iterates through all embedding modules in the model and assures it can embed
        with the extended vocab. This is required in fine-tuning or transfer learning
        scenarios where model was trained with original vocabulary but during
        fine-tuning/transfer-learning, it will have it work with extended vocabulary
        (original + new-data vocabulary).

        # Parameters

        embedding_sources_mapping : `Dict[str, str]`, optional (default = `None`)
            Mapping from model_path to pretrained-file path of the embedding
            modules. If pretrained-file used at time of embedding initialization
            isn't available now, user should pass this mapping. Model path is
            path traversing the model attributes upto this embedding module.
            Eg. "_text_field_embedder.token_embedder_tokens".
        """
        # self.named_modules() gives all sub-modules (including nested children)
        # The path nesting is already separated by ".": eg. parent_module_name.child_module_name
        embedding_sources_mapping = embedding_sources_mapping or {}
        for model_path, module in self.named_modules():
            if hasattr(module, "extend_vocab"):
                pretrained_file = embedding_sources_mapping.get(model_path)
                # Show useful information when reading from a pretrained file, kind of an ugly hack
                if module._pretrained_file is not None:
                    original_logging_level = logging.getLogger("allennlp").level
                    logging.getLogger("allennlp").setLevel("INFO")

                module.extend_vocab(
                    self.vocab,
                    extension_pretrained_file=pretrained_file,
                    model_path=model_path,
                )

                if module._pretrained_file is not None:
                    logging.getLogger("allennlp").setLevel(original_logging_level)

    @property
    def inputs(self) -> List[str]:
        """The model inputs. Corresponding to head.featurize required argument names"""
        return self._inputs

    @property
    def output(self) -> List[str]:
        """The model outputs (not prediction): Corresponding to head.featurize optional argument names."""
        return self._output

    def init_prediction_logger(
        self, output_dir: str, max_bytes: int = 20000000, backup_count: int = 20
    ):
        """Initialize the prediction logger.

        If initialized we will log all predictions to a file called *predictions.json* in the `output_folder`.

        Parameters
        ----------
        output_dir
            Path to the folder in which we create the *predictions.json* file.
        max_bytes
            Passed on to logging.handlers.RotatingFileHandler
        backup_count
            Passed on to logging.handlers.RotatingFileHandler

        """
        Path(output_dir).mkdir(parents=True, exist_ok=True)

        predictions_logger = logging.getLogger(output_dir)
        predictions_logger.setLevel(logging.DEBUG)
        # This flag avoids logging messages to be propagated to the parent loggers
        predictions_logger.propagate = False
        file_handler = RotatingFileHandler(
            os.path.join(output_dir, self.PREDICTION_FILE_NAME),
            maxBytes=max_bytes,
            backupCount=backup_count,
        )
        file_handler.setLevel(logging.INFO)
        predictions_logger.addHandler(file_handler)

        setattr(self, "_prediction_logger", predictions_logger)

    def init_prediction_cache(self, max_size: int) -> None:
        """Initialize a prediction cache using the functools.lru_cache decorator

        Parameters
        ----------
        max_size
            Save up to max_size most recent items.
        """

        if hasattr(self, "_predict_with_cache"):
            warnings.warn(
                "Prediction cache already initiated!", category=RuntimeWarning
            )
            return

        predict_with_cache = lru_cache(maxsize=max_size)(self.predict)

        def predict_wrapper(*args, **kwargs):
            def hashable_value(value) -> Any:
                if isinstance(value, dict):
                    return _HashDict(value)
                if isinstance(value, (list, tuple)):
                    return _HashList(value)
                return value

            return predict_with_cache(
                *[hashable_value(arg_value) for arg_value in args],
                **{
                    key: hashable_value(input_value)
                    for key, input_value in kwargs.items()
                },
            )

        self.__setattr__("predict", predict_wrapper)
        self.__setattr__("_predict_with_cache", predict_with_cache)

    def _log_predictions(
        self,
        batch: Iterable[Dict[str, Any]],
        predictions: Iterable[TaskPrediction],
    ) -> None:
        """Log predictions to a file for a model analysis and feedback sessions.

        Parameters
        ----------
        batch
            Input data to the model
        predictions
            Returned predictions from the model
        """
        for input_dict, prediction in zip(batch, predictions):
            self._prediction_logger.info(
                json.dumps(
                    dict(inputs=input_dict, prediction=sanitize(prediction.as_dict()))
                )
            )

    def predict(
        self,
        batch: List[Dict[str, Union[str, List[str], Dict[str, str]]]],
        prediction_config: PredictionConfiguration,
    ) -> List[Optional[TaskPrediction]]:
        """Returns predictions given some input data based on the current state of the model

        The keys of the input dicts in the batch must coincide with the `self.inputs` attribute.
        TODO: Comply with LightningModule API + Trainer API (means move instance creation logic to Pipeline)

        Parameters
        ----------
        batch
            A list of dictionaries that represents a batch of inputs.
            The dictionary keys must comply with the `self.inputs` attribute.
        prediction_config
            Contains configurations for the prediction
        """
        if self.training:
            self.eval()

        instances = [self.text_to_instance(**input_dict) for input_dict in batch]
        # Filter out None instances, that is when the head could not create an instance out of the input
        none_indices, not_none_instances = [], []
        for i, instance in enumerate(instances):
            if instance is None:
                none_indices.append(i)
            else:
                not_none_instances.append(instance)

        try:
            forward_outputs = self.forward_on_instances(not_none_instances)
        except Exception as error:
            input_examples = [
                example for i, example in enumerate(batch) if i not in none_indices
            ]
            self._LOGGER.exception(error)
            self._LOGGER.warning(
                f"Failed to make a forward pass for '{input_examples}'"
            )
            return [None] * len(batch)

        predictions = []
        for forward_output, instance in zip(forward_outputs, not_none_instances):
            try:
                predictions.append(
                    self.head.make_task_prediction(
                        forward_output, instance, prediction_config
                    )
                )
            except Exception as error:
                self._LOGGER.exception(error)
                self._LOGGER.warning(
                    f"Failed to make a task prediction for '{forward_output, instance}'"
                )
                predictions.append(None)

        # Add None for the none instances
        for index in none_indices:
            predictions.insert(index, None)

        # Log predictions if the prediction logger was initialized
        if hasattr(self, "_prediction_logger"):
            self._log_predictions(batch, predictions)

        return predictions

    def on_fit_start(self) -> None:
        # Reset metrics
        self.best_metrics = None

    def training_step(self, batch, batch_idx) -> Dict:
        output = self(**batch)
        self.log(
            "training_loss",
            output["loss"],
            on_step=True,
            prog_bar=False,
            on_epoch=False,
        )

        metrics = self.get_metrics()
        for key, val in metrics.items():
            self.log(
                ("training" if key.startswith("_") else "training_") + key,
                val,
                on_step=True,
                prog_bar=not key.startswith("_"),
                on_epoch=False,
            )

        return output

    def training_epoch_end(self, outputs: List[Any]) -> None:
        metrics = self.get_metrics(reset=True)
        for key, val in metrics.items():
            if key.startswith("_"):
                metric_name = self.TRAINING_METRICS_PREFIX + key
            else:
                metric_name = self.TRAINING_METRICS_PREFIX + "_" + key
            self.log(
                metric_name,
                val,
                on_step=False,
                prog_bar=False,
                on_epoch=True,
            )

    def validation_step(self, batch, batch_idx) -> Dict:
        output = self(**batch)
        self.get_metrics()

        return output

    def validation_epoch_end(self, outputs: List[Any]) -> None:
        # we do not want to log any metrics for the sanity check
        if self.trainer.sanity_checking:
            return

        # we keep track of the logged metrics to figure out the best metrics
        logged_metrics = {}

        averaged_epoch_loss = sum([output["loss"] for output in outputs]) / len(outputs)
        self.log(
            f"{self.VALIDATION_METRICS_PREFIX}_loss",
            averaged_epoch_loss,
            on_step=False,
            prog_bar=True,
            on_epoch=True,
        )
        logged_metrics[f"{self.VALIDATION_METRICS_PREFIX}_loss"] = averaged_epoch_loss

        metrics = self.get_metrics(reset=True)
        for key, val in metrics.items():
            if key.startswith("_"):
                metric_name = self.VALIDATION_METRICS_PREFIX + key
            else:
                metric_name = self.VALIDATION_METRICS_PREFIX + "_" + key
            self.log(
                metric_name,
                val,
                on_step=False,
                prog_bar=not key.startswith("_"),
                on_epoch=True,
            )
            logged_metrics[metric_name] = val

        # log best metrics
        logged_metrics["epoch"] = self.current_epoch
        if self.best_metrics is None:
            self.best_metrics = logged_metrics
        elif (
            self.monitor_mode == "max"
            and self.best_metrics[self.monitor] < logged_metrics[self.monitor]
        ):
            self.best_metrics = logged_metrics
        elif (
            self.monitor_mode == "min"
            and self.best_metrics[self.monitor] > logged_metrics[self.monitor]
        ):
            self.best_metrics = logged_metrics

        self.log_dict(
            {f"best_{key}": value for key, value in self.best_metrics.items()},
            on_step=False,
            prog_bar=False,
            on_epoch=True,
        )

    def test_step(self, *args, **kwargs) -> Dict:
        return self.validation_step(*args, **kwargs)

    def test_epoch_end(self, outputs: List[Any]) -> None:
        averaged_epoch_loss = sum([output["loss"] for output in outputs]) / len(outputs)
        self.log(
            f"{self.TEST_METRICS_PREFIX}_loss",
            averaged_epoch_loss,
            on_step=False,
            prog_bar=True,
            on_epoch=True,
        )

        metrics = self.get_metrics(reset=True)
        for key, val in metrics.items():
            if key.startswith("_"):
                metric_name = self.TEST_METRICS_PREFIX + key
            else:
                metric_name = self.TEST_METRICS_PREFIX + "_" + key
            self.log(
                metric_name,
                val,
                on_step=False,
                prog_bar=not key.startswith("_"),
                on_epoch=True,
            )

    def configure_optimizers(self):
        if self.lr_scheduler is None:
            return self.optimizer
        return [self.optimizer], [self.lr_scheduler]


allennlp.models.Model.register(PipelineModel.__name__, exist_ok=True)(PipelineModel)



import pytest
import yaml
from allennlp.common.checks import ConfigurationError
from allennlp.data.fields import ListField
from allennlp.data.fields import TextField
from spacy.tokens.token import Token

from biome.text import Pipeline
from biome.text.configuration import CharFeatures
from biome.text.configuration import FeaturesConfiguration
from biome.text.configuration import PipelineConfiguration
from biome.text.configuration import TaskHeadConfiguration
from biome.text.configuration import TokenizerConfiguration
from biome.text.configuration import WordFeatures
from biome.text.modules.configuration.allennlp_configuration import (
    Seq2SeqEncoderConfiguration,
)
from biome.text.modules.heads import TextClassification


@pytest.fixture
def pipeline_yaml(tmp_path):
    pipeline_dict = {
        "name": "test_pipeline_config",
        "tokenizer": {
            "text_cleaning": {"rules": ["strip_spaces"]},
            "use_spacy_tokens": True,
        },
        "features": {
            "word": {"embedding_dim": 2, "lowercase_tokens": True},
            "char": {
                "embedding_dim": 2,
                "encoder": {
                    "type": "gru",
                    "hidden_size": 2,
                    "num_layers": 1,
                    "bidirectional": True,
                },
                "dropout": 0.1,
            },
        },
        "encoder": {
            "type": "gru",
            "hidden_size": 2,
            "num_layers": 1,
            "bidirectional": True,
        },
        "head": {
            "type": "TextClassification",
            "labels": ["duplicate", "not_duplicate"],
            "pooler": {"type": "boe"},
        },
    }

    yaml_path = tmp_path / "pipeline.yml"
    with yaml_path.open("w") as file:
        yaml.safe_dump(pipeline_dict, file)

    return str(yaml_path)


@pytest.fixture
def transformers_pipeline_config():
    return {
        "name": "transformers_tokenizer_plus_tokenclassification",
        "features": {
            "transformers": {"model_name": "sshleifer/tiny-distilroberta-base"}
        },
        "head": {
            "type": "TextClassification",
            "labels": ["duplicate", "not_duplicate"],
        },
    }


def test_pipeline_without_word_features():
    tokenizer_config = TokenizerConfiguration()
    char_features = CharFeatures(
        embedding_dim=2,
        encoder={
            "type": "gru",
            "hidden_size": 2,
            "num_layers": 1,
            "bidirectional": True,
        },
        dropout=0.1,
    )
    features_config = FeaturesConfiguration(char=char_features)
    encoder_spec = Seq2SeqEncoderConfiguration(
        type="gru", hidden_size=2, num_layers=1, bidirectional=True
    )

    head_spec = TaskHeadConfiguration(
        type="TextClassification",
        labels=["duplicate", "not_duplicate"],
        pooler={"type": "boe"},
    )

    pipeline_config = PipelineConfiguration(
        name="no_word_features",
        head=head_spec,
        features=features_config,
        tokenizer=tokenizer_config,
        encoder=encoder_spec,
    )

    pl = Pipeline.from_config(pipeline_config)
    assert "word" not in pl.backbone.featurizer.indexer
    assert "char" in pl.backbone.featurizer.indexer


def test_pipeline_config(pipeline_yaml):
    tokenizer_config = TokenizerConfiguration(
        text_cleaning={"rules": ["strip_spaces"]}, use_spacy_tokens=True
    )

    word_features = WordFeatures(embedding_dim=2, lowercase_tokens=True)
    char_features = CharFeatures(
        embedding_dim=2,
        encoder={
            "type": "gru",
            "hidden_size": 2,
            "num_layers": 1,
            "bidirectional": True,
        },
        dropout=0.1,
    )
    features_config = FeaturesConfiguration(word=word_features, char=char_features)

    encoder_spec = Seq2SeqEncoderConfiguration(
        type="gru", hidden_size=2, num_layers=1, bidirectional=True
    )

    head_spec = TaskHeadConfiguration(
        type=TextClassification,
        labels=["duplicate", "not_duplicate"],
        pooler={"type": "boe"},
    )

    pipeline_config = PipelineConfiguration(
        name="test_pipeline_config",
        head=head_spec,
        features=features_config,
        tokenizer=tokenizer_config,
        encoder=encoder_spec,
    )

    pl = Pipeline.from_config(pipeline_config)

    pl_yaml = Pipeline.from_yaml(pipeline_yaml)

    assert pl.named_trainable_parameters == pl_yaml.named_trainable_parameters
    assert pl.num_trainable_parameters == pl_yaml.num_trainable_parameters
    assert pl.num_parameters == pl_yaml.num_parameters

    sample_text = "My simple text"
    for instance in [
        pl.backbone.featurizer(sample_text),
        pl_yaml.backbone.featurizer(sample_text),
    ]:
        for key, value in instance.items():
            assert key == "record"
            assert isinstance(value, ListField)
            assert len(value) == 1
            for text in value:
                assert isinstance(text, TextField)
                assert all(map(lambda t: isinstance(t, Token), text.tokens))
                assert sample_text == " ".join([t.text for t in text.tokens])


def test_invalid_tokenizer_features_combination(transformers_pipeline_config):
    transformers_pipeline_config["features"].update({"word": {"embedding_dim": 2}})
    transformers_pipeline_config["tokenizer"] = {"use_transformers": True}

    with pytest.raises(ConfigurationError):
        Pipeline.from_config(transformers_pipeline_config)


def test_not_implemented_transformers_with_tokenclassification(
    transformers_pipeline_config,
):
    transformers_pipeline_config["tokenizer"] = {"use_transformers": True}
    transformers_pipeline_config["head"] = {
        "type": "TokenClassification",
        "labels": ["NER"],
    }
    with pytest.raises(NotImplementedError):
        Pipeline.from_config(transformers_pipeline_config)


def test_invalid_transformers_tokenizer_indexer_embedder_combination(
    transformers_pipeline_config,
):
    transformers_pipeline_config["tokenizer"] = {
        "transformers_kwargs": {"model_name": "distilroberta-base"}
    }

    with pytest.raises(ConfigurationError):
        Pipeline.from_config(transformers_pipeline_config)



from typing import Tuple

import pytest
from pytorch_lightning import seed_everything

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import VocabularyConfiguration
from biome.text.configuration import CharFeatures
from biome.text.configuration import TrainerConfiguration
from biome.text.configuration import WordFeatures


@pytest.fixture
def train_valid_dataset(resources_data_path) -> Tuple[Dataset, Dataset]:
    """Returns both training and validation datasets"""

    training_ds = Dataset.from_csv(
        paths=str(resources_data_path / "business.cat.2k.train.csv")
    )
    validation_ds = Dataset.from_csv(
        paths=str(resources_data_path / "business.cat.2k.valid.csv")
    )

    return training_ds, validation_ds


@pytest.fixture
def pipeline_dict() -> dict:
    return {
        "name": "german_business_names",
        "features": {
            "word": {"embedding_dim": 16, "lowercase_tokens": True},
            "char": {
                "embedding_dim": 16,
                "encoder": {
                    "type": "gru",
                    "num_layers": 1,
                    "hidden_size": 32,
                    "bidirectional": True,
                },
                "dropout": 0.1,
            },
        },
        "head": {
            "type": "TextClassification",
            "labels": [
                "Unternehmensberatungen",
                "Friseure",
                "Tiefbau",
                "Dienstleistungen",
                "Gebrauchtwagen",
                "Restaurants",
                "Architekturbüros",
                "Elektriker",
                "Vereine",
                "Versicherungsvermittler",
                "Sanitärinstallationen",
                "Edv",
                "Maler",
                "Physiotherapie",
                "Werbeagenturen",
                "Apotheken",
                "Vermittlungen",
                "Hotels",
                "Autowerkstätten",
                "Elektrotechnik",
                "Allgemeinärzte",
                "Handelsvermittler Und -vertreter",
            ],
            "pooler": {
                "type": "gru",
                "num_layers": 1,
                "hidden_size": 16,
                "bidirectional": True,
            },
            "feedforward": {
                "num_layers": 1,
                "hidden_dims": [16],
                "activations": ["relu"],
                "dropout": [0.1],
            },
        },
    }


def test_text_classification(tmp_path, pipeline_dict, train_valid_dataset):
    """Apart from a well specified training, this also tests the vocab creation!"""
    seed_everything(43)

    pl = Pipeline.from_config(pipeline_dict)
    train_ds = train_valid_dataset[0]
    valid_ds = train_valid_dataset[1]

    vocab_config = VocabularyConfiguration(max_vocab_size={"word": 50})
    trainer_config = TrainerConfiguration(
        batch_size=64,
        optimizer={"type": "adam", "lr": 0.01},
        max_epochs=5,
        default_root_dir=str(tmp_path),
        gpus=0,  # turn off gpus even if available
    )

    trainer = Trainer(
        pipeline=pl,
        train_dataset=train_ds,
        valid_dataset=valid_ds,
        trainer_config=trainer_config,
        vocab_config=vocab_config,
    )

    trainer.fit(tmp_path / "output")

    assert pl.vocab.get_vocab_size(WordFeatures.namespace) == 52
    assert pl.vocab.get_vocab_size(CharFeatures.namespace) == 83

    assert pl.num_trainable_parameters == 22070

    evaluation = trainer.test(valid_ds, batch_size=16)

    # Reminder: the value depends on the batch_size!
    assert evaluation["test_loss"] == pytest.approx(0.7404146790504456, abs=0.003)

    Pipeline.from_pretrained(str(tmp_path / "output" / "model.tar.gz"))

    assert pl.vocab.get_vocab_size(WordFeatures.namespace) == 52
    assert pl.vocab.get_vocab_size(CharFeatures.namespace) == 83



import pytest
from numpy.testing import assert_allclose

from biome.text import Pipeline


@pytest.fixture
def pipeline():
    return Pipeline.from_config(
        {
            "name": "test_pipeline_copy",
            "head": {"type": "TextClassification", "labels": ["a", "b"]},
        }
    )


def test_save(pipeline, tmp_path):
    pipeline.save(tmp_path)

    assert (tmp_path / "model.tar.gz").is_file()

    expected_prediction = pipeline.predict("test")
    prediction = Pipeline.from_pretrained(tmp_path / "model.tar.gz").predict("test")

    assert prediction["labels"] == expected_prediction["labels"]
    assert_allclose(prediction["probabilities"], expected_prediction["probabilities"])



import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

import numpy
import torch
from allennlp.common.checks import ConfigurationError
from allennlp.data import Instance
from allennlp.data import TextFieldTensors
from allennlp.modules import SoftmaxLoss
from allennlp.nn.util import get_text_field_mask
from allennlp.nn.util import get_token_ids_from_text_field_tensors

from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.metrics import Metrics
from biome.text.modules.configuration import ComponentConfiguration
from biome.text.modules.heads.task_head import TaskHead
from biome.text.modules.heads.task_head import TaskName
from biome.text.modules.heads.task_prediction import LanguageModellingPrediction


class LanguageModelling(TaskHead):
    """
    Task head for next-token language modelling, i.e., a model to predict the next token
    in a sequence of tokens.
    """

    task_name = TaskName.language_modelling
    _LOGGER = logging.getLogger(__name__)

    def __init__(
        self,
        backbone: ModelBackbone,
        dropout: float = None,
        bidirectional: bool = False,
    ) -> None:
        super(LanguageModelling, self).__init__(backbone)

        self.bidirectional = bidirectional

        if not backbone.featurizer.has_word_features:
            raise ConfigurationError(
                "`LanguageModelling` defines a word-level next token language model. "
                "Please check your `features` configuration to enable at least `words` features."
            )

        if backbone.encoder.is_bidirectional() is not bidirectional:
            raise ConfigurationError(
                "Bidirectionality of contextualizer must match bidirectionality of "
                "language model. "
                f"Contextualizer bidirectional: {backbone.encoder.is_bidirectional()}, "
                f"language model bidirectional: {bidirectional}"
            )

        if self.bidirectional:
            self._forward_dim = backbone.encoder.get_output_dim() // 2
        else:
            self._forward_dim = backbone.encoder.get_output_dim()

        if dropout:
            self._dropout = torch.nn.Dropout(dropout)
        else:
            self._dropout = lambda x: x

        self._metrics = Metrics(perplexity={"type": "perplexity"})

        self._loss = SoftmaxLoss(
            num_words=vocabulary.words_vocab_size(self.backbone.vocab),
            embedding_dim=self._forward_dim,
        )

    def on_vocab_update(self):
        num_words = vocabulary.words_vocab_size(self.backbone.vocab)
        if len(self._loss.softmax_b) != num_words:
            self._loss = SoftmaxLoss(
                num_words=num_words,
                embedding_dim=self._forward_dim,
            )

    def featurize(self, text: str) -> Optional[Instance]:
        instance = self.backbone.featurizer(text, to_field="text", aggregate=True)

        return instance

    def forward(self, text: TextFieldTensors) -> Dict[str, Any]:  # type: ignore

        mask = get_text_field_mask(text)
        contextual_embeddings = self.backbone.forward(text, mask)

        token_ids = get_token_ids_from_text_field_tensors(text)
        assert isinstance(contextual_embeddings, torch.Tensor)

        # Use token_ids to compute targets
        # targets are next token ids with respect to first token in the seq
        # e.g. token_ids [[1, 3, 5, 7],..[]], forward_targets=[[3,5,7],..]
        forward_targets = torch.zeros_like(token_ids)
        forward_targets[:, 0:-1] = token_ids[:, 1:]

        if self.bidirectional:
            backward_targets = torch.zeros_like(token_ids)
            backward_targets[:, 1:] = token_ids[:, 0:-1]
        else:
            backward_targets = None

        # add dropout
        contextual_embeddings_with_dropout = self._dropout(contextual_embeddings)

        # compute softmax loss
        try:
            forward_loss, backward_loss = self._compute_loss(
                contextual_embeddings_with_dropout, forward_targets, backward_targets
            )
        except IndexError:
            raise IndexError(
                "Word token out of vocabulary boundaries, please check your vocab is correctly set"
                " or created before starting training."
            )

        num_targets = torch.sum((forward_targets > 0).long())

        if num_targets > 0:
            if self.bidirectional:
                average_loss = (
                    0.5 * (forward_loss + backward_loss) / num_targets.float()
                )
            else:
                average_loss = forward_loss / num_targets.float()
        else:
            average_loss = torch.tensor(0.0)

        for metric in self._metrics.get_dict(is_train=self.training).values():
            # Perplexity needs the value to be on the cpu
            metric(average_loss.to("cpu"))

        return dict(
            loss=average_loss,
            lm_embeddings=contextual_embeddings,
            mask=mask,
        )

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        return {
            metric_name: metric.get_metric(reset)
            for metric_name, metric in self._metrics.get_dict(
                is_train=self.training
            ).items()
        }

    def _compute_loss(
        self,
        lm_embeddings: torch.Tensor,
        forward_targets: torch.Tensor,
        backward_targets: torch.Tensor = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # If bidirectional, lm_embeddings is shape (batch_size, timesteps, dim * 2)
        # If unidirectional, lm_embeddings is shape (batch_size, timesteps, dim)
        # forward_targets, backward_targets (None in the unidirectional case) are
        # shape (batch_size, timesteps) masked with 0
        if self.bidirectional:
            forward_embeddings, backward_embeddings = lm_embeddings.chunk(2, -1)
            backward_loss = self._loss_helper(backward_embeddings, backward_targets)
        else:
            forward_embeddings = lm_embeddings
            backward_loss = None

        forward_loss = self._loss_helper(forward_embeddings, forward_targets)
        return forward_loss, backward_loss

    def _loss_helper(
        self,
        direction_embeddings: torch.Tensor,
        direction_targets: torch.Tensor,
    ) -> torch.Tensor:
        mask = direction_targets > 0
        # we need to subtract 1 to undo the padding id since the softmax
        # does not include a padding dimension

        # shape (batch_size * timesteps, )
        non_masked_targets = direction_targets.masked_select(mask) - 1

        # shape (batch_size * timesteps, embedding_dim)
        non_masked_embeddings = direction_embeddings.masked_select(
            mask.unsqueeze(-1)
        ).view(-1, self._forward_dim)

        return self._loss(non_masked_embeddings, non_masked_targets)

    def _make_task_prediction(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
    ) -> LanguageModellingPrediction:
        task_prediction = LanguageModellingPrediction(
            lm_embeddings=single_forward_output["lm_embeddings"],
            mask=single_forward_output["mask"],
        )
        if "loss" in single_forward_output:
            task_prediction.loss = float(single_forward_output["loss"])

        return task_prediction


class LanguageModellingConfiguration(ComponentConfiguration[LanguageModelling]):
    """Configuration for language model head components"""

    pass



import logging
import os

import pytest
import torch

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import PipelineConfiguration
from biome.text import Trainer
from biome.text import TrainerConfiguration
from biome.text.backbone import ModelBackbone
from biome.text.modules.heads import TextClassification
from biome.text.modules.heads import TextClassificationConfiguration


class TestHead(TextClassification):
    def __init__(self, backbone: ModelBackbone):
        super(TestHead, self).__init__(backbone, labels=["test", "notest"])


@pytest.fixture
def dataset(tmp_path) -> Dataset:
    data = {
        "text": ["A common text", "This is why you get", "Seriosly?, I'm not sure"],
        "label": ["one", "zero", "zero"],
    }
    ds = Dataset.from_dict(data)

    # we save and load it here to be able to lazily read from it
    ds_path = tmp_path / "test_pipeline_datasets" / "dataset"
    ds.save_to_disk(str(ds_path))

    return Dataset.load_from_disk(str(ds_path))


@pytest.fixture
def pipeline() -> Pipeline:
    config = PipelineConfiguration(
        name="test-classifier",
        head=TextClassificationConfiguration(labels=["one", "zero"]),
    )
    return Pipeline.from_config(config)


def test_training_from_pretrained_with_head_replace(pipeline, dataset, tmp_path):
    trainer_config = TrainerConfiguration(
        batch_size=2,
        max_epochs=5,
        gpus=0,
    )

    trainer = Trainer(pipeline, train_dataset=dataset, trainer_config=trainer_config)
    trainer.fit(tmp_path / "output")

    pipeline.set_head(TestHead)
    pipeline.config.tokenizer_config.max_nr_of_sentences = 3
    copied = pipeline.copy()
    assert isinstance(copied.head, TestHead)
    assert copied.num_parameters == pipeline.num_parameters
    assert copied.num_trainable_parameters == pipeline.num_trainable_parameters
    copied_model_state = copied._model.state_dict()
    original_model_state = pipeline._model.state_dict()
    for key, value in copied_model_state.items():
        if "backbone" in key:
            assert torch.all(torch.eq(value, original_model_state[key]))
    assert copied.backbone.featurizer.tokenizer.config.max_nr_of_sentences == 3



# fmt: off
from .allennlp_configuration import BiMpmMatchingConfiguration
from .allennlp_configuration import EmbeddingConfiguration
from .allennlp_configuration import FeedForwardConfiguration
from .allennlp_configuration import Seq2SeqEncoderConfiguration
from .allennlp_configuration import Seq2VecEncoderConfiguration
from .defs import ComponentConfiguration

# fmt: on
